from model_convertor import convert_model

from typing import Dict

from pydtnn.activations.softmax import Softmax
from pydtnn.datasets.dataset import select as select_dataset
from torchvision.models import vgg19, alexnet, densenet169, resnet50, googlenet
from torchvision.models import densenet121, densenet201, resnet18, resnet34, resnet101, resnet152, vgg11, vgg16

from torchmetrics import Accuracy, Metric

from pydtnn.datasets.dataset import Dataset

from pydtnn.models.vgg11 import vgg11 as pydtnn_vgg11
from pydtnn.models.vgg16 import vgg16 as pydtnn_vgg16
from pydtnn.models.vgg19_imagenet import vgg19_imagenet as pydtnn_vgg19_imagenet
from pydtnn.models.alexnet_cifar10 import alexnet_cifar10 as pydtnn_alexnet_cifar10
from pydtnn.models.densenet121_cifar10 import densenet121_cifar10 as pydtnn_densenet121_cifar10
from pydtnn.models.densenet169_cifar10 import densenet169_cifar10 as pydtnn_densenet169_cifar10
from pydtnn.models.densenet201_cifar10 import densenet201_cifar10 as pydtnn_densenet201_cifar10
from pydtnn.models.resnet18_cifar10 import resnet18_cifar10 as pydtnn_resnet18_cifar10
from pydtnn.models.resnet34_cifar10 import resnet34_cifar10 as pydtnn_resnet34_cifar10
from pydtnn.models.resnet50_cifar10 import resnet50_cifar10 as pydtnn_resnet50_cifar10
from pydtnn.models.resnet101_cifar10 import resnet101_cifar10 as pydtnn_resnet101_cifar10
from pydtnn.models.resnet152_cifar10 import resnet152_cifar10 as pydtnn_resnet152_cifar10
from pydtnn.models.inceptionv3_cifar10 import inceptionv3_cifar10 as pydtnn_inceptionv3_cifar10

from pydtnn.model import Model as PyDTNN_Model
from pydtnn.datasets.dataset import select as select_dataset
from pydtnn.utils.best_of import BestOf

import torch
from torch.nn import CrossEntropyLoss


dict_test = {
    "vgg11": (vgg11, pydtnn_vgg11, (524, 524, 3), "cifar10", {"num_classes": 5}, None),  # (224, 224, 3)
    "vgg16": (vgg16, pydtnn_vgg16, (524, 524, 3), "cifar10", {"num_classes": 5}, None),  # (224, 224, 3)
    "vgg19": (vgg19, pydtnn_vgg19_imagenet, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "alexnet": (alexnet, pydtnn_alexnet_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, None),
    "densenet121": (densenet121, pydtnn_densenet121_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "densenet169": (densenet169, pydtnn_densenet169_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "densenet201": (densenet201, pydtnn_densenet201_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "resnet18": (resnet18, pydtnn_resnet18_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, None),
    "resnet34": (resnet34, pydtnn_resnet34_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, None),
    "resnet50": (resnet50, pydtnn_resnet50_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "resnet101": (resnet101, pydtnn_resnet101_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, None),
    "resnet152": (resnet152, pydtnn_resnet152_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, not None),
    "googlenet": (googlenet, pydtnn_inceptionv3_cifar10, (524, 524, 3), "cifar10", {"num_classes": 5}, None),  # (299, 299, 3)
}

# ----- EXECUTION PARAMETERS ----- #
TEST = "densenet169"
FIRST_PYTORCH = False
OLD_FIRST = None
DATASET_PATH = ""
WEIGHTS_PATH = ""
OUTPUT_PATH = ""
INFERENCE = True
TRAINING = False

KWARGS = {
    "model_name": None,
    "dataset": None,
    "dataset_name": None,
    "evaluate_only": True,
    "parallel": "data",
    "tensor_format": "NCHW",  # "NCHW" # "NHWC",
    "loss_func": "categorical_cross_entropy",
    "enable_gpu": False,  # True,
    "dataset_train_path": DATASET_PATH,
    "dataset_test_path": DATASET_PATH,
}


def get_model_layers(model: torch.nn.Module, name: str = "self") -> Dict[str, torch.nn.Module]:
    # Recursive function to get the models without containers modules.
    def _get_model_layers(model: torch.nn.Module, name: str, dict_modules: Dict[str, torch.nn.Module]):
        # The recursive function.
        children = list(model.named_children())
        if len(children) > 0:
            for nom, module in children:
                _get_model_layers(model=module, name=".".join([name, nom]), dict_modules=dict_modules)
        else:
            dict_modules[name] = model
    # -- END _get_model_layers--#
    dict_modules = {}
    _get_model_layers(model=model, name=name, dict_modules=dict_modules)
    return dict_modules


def pytorch_inference(model: torch.nn.Module, dataloader, loss_func: torch.nn.modules.loss._Loss, device: torch.device, metrics_list: list) -> None:

    outputs_list = list()
    labels_list = list()

    print(f"metrics_list: {metrics_list}")

    if False:
        dict_layers = get_model_layers(model)
        for n in dict_layers.keys():
            m = dict_layers[n]

            def print_pre(module, x, *, name=n):
                print(f"Layer - {name}:")
                if isinstance(x[0], list):
                    for elem in x[0]:
                        print(f"\telem.size(): {elem.size()}")
                else:
                    print(f"\tx[0].size(): {x[0].size()}")
            # ----

            def print_post(module, args, output, *, name=n):
                print(f"{name}:")
                # print(f"output: {output}")
                if isinstance(output[0], list):
                    for elem in output[0]:
                        print(f"\toutput.size(): {elem.size()}")
                        # print(f"\toutput: {elem}")
                else:
                    print(f"\toutput[0].size(): {output[0].size()}")
                    # print(f"\toutput[0]: {output[0]}")
            # ----

            m.register_forward_pre_hook(print_pre)
            m.register_forward_hook(print_post)
        # --------

    model.eval()
    with torch.no_grad():

        for inputs, labels, _ in dataloader:
            inputs = torch.Tensor(inputs).to(device)
            labels = torch.Tensor(labels).to(device)
            outputs = model(inputs)
            outputs = outputs.to(device)
            # loss = loss_fn(outputs, labels)

            # print(f"outputs:\n{outputs}")
            outputs_list.extend(outputs)
            labels_list.extend(labels)
            for _, metric in metrics_list:
                metric: Metric = metric
                metric.update(outputs, labels)

    print(f"metrics_list: {metrics_list}")

    for name, metric in metrics_list:
        print(f"type(metric): {type(metric)}")
        metric_result = metric.compute()
        print(f"{name}: {metric_result:.4f}")


def print_model_reports(model):
    # Print performance counter report
    model.perf_counter.print_report()
    # Print BestOf report
    if model.enable_best_of:
        print()
        BestOf.print_report()


def pydtnn_inference(model: PyDTNN_Model, metrics_list=None, dataset=None) -> None:
    metrics_list = [f for f in model.metrics.replace(" ", "").split(",")] if metrics_list is None else metrics_list
    model.dataset = dataset if dataset is not None else model.dataset
    model.show()
    model.evaluate_dataset()
    print_model_reports(model)


def _pydtnn_inference(new_model, old_model, dataset, old_first=None):
    print("-------------------")
    print(" PyDTNN's inference")
    print("-------------------")

    match old_first:
        case True:
            print("OLD model")
            pydtnn_inference(model=old_model, dataset=dataset)
            print("NEW model")
            pydtnn_inference(model=new_model, dataset=dataset)
        case False:
            print("NEW model")
            pydtnn_inference(model=new_model, dataset=dataset)
            print("OLD model")
            pydtnn_inference(model=old_model, dataset=dataset)
        case not_old_model:  # old_first = None
            print("NEW model")
            pydtnn_inference(model=new_model, dataset=dataset)
# -----------------------#


def _pytorch_inference(pytorch_model, dataloader, kwargs, device):
    print("-------------------")
    print("Pytorch's inference")
    print("-------------------")

    match kwargs["loss_func"]:
        case "categorical_cross_entropy":
            loss = CrossEntropyLoss()
        case _:
            loss = None
            print("Pick another loss")
            assert False

    task = "binary"
    num_classes = 10
    pytorch_inference(model=pytorch_model, dataloader=dataloader, loss_func=loss, device=device,
                      metrics_list=[("Accuracy", Accuracy(task=task, num_classes=num_classes)),
                                    # ("AUROC", AUROC(task = task, num_classes = num_classes)),
                                    # ("AveragePrecision", AveragePrecision(task = task, num_classes = num_classes)),
                                    # ("F1Score", F1Score(task = task, num_classes = num_classes))
                                    ])
# -----------------------#


def pydtnn_training(model: PyDTNN_Model, dataset: Dataset, num_samples=64 * 2):

    # history = model.train(x_train=dataset._x[DatasetEnum.TRAIN][:num_samples], x_val=dataset._x[VAL][:num_samples],
    #                      y_train=dataset._y[DatasetEnum.TRAIN][:num_samples], y_val=dataset._y[VAL][:num_samples])
    history = model.train_dataset()
    print(f"history: {history}")


def main():
    test = TEST

    pytorch_model, create_pydtnn_model, shape, dataset, args, weight = dict_test[test]
    pytorch_model: torch.Module = pytorch_model(**args)

    KWARGS["dataset"] = dataset
    KWARGS["dataset_name"] = dataset
    output_shape = args["num_classes"]
    kwargs = KWARGS
    shape = shape if KWARGS["tensor_format"] == "NHWC" else (shape[2], *shape[:2])
    print(f"{shape}")

    device = torch.device("cpu")  # if kwargs["enable_gpu"] == False else torch.device("cuda")
    if weight is not None:
        weight = f"{WEIGHTS_PATH}model_{test}.pth"
        weight = torch.load(weight, weights_only=True, map_location=torch.device(device))

        pytorch_model.load_state_dict(weight, strict=False,)

    print("====================")
    print("== PyDTNN version ==")
    print("====================")

    old_model = PyDTNN_Model(**kwargs)
    create_pydtnn_model(old_model, output_shape=output_shape)
    print("PyDTNN version:")
    old_model.show()
    print("-----\n")

    print("=====================")
    print("== PyTorch version ==")
    print("=====================")
    print(pytorch_model)
    print("-----\n")

    print("PyTorch model's forward method:")
    graph = torch.fx.symbolic_trace(pytorch_model)
    print(graph.code)

    print("-----\n")

    # dataset = get_dataset(old_model)
    # pydtnn_inference(model=old_model, dataset = dataset)

    print("=======================")
    print("== Converted version ==")
    print("=======================")

    new_model = convert_model(model=pytorch_model, input_shape=shape,
                              default_output_activation_layer=Softmax(), **kwargs)

    print("=====================")
    print("=== MODEL CREATED ===")
    print("=====================")

    new_model.show()
    print("-----")
    print(f"type(new_model): {type(new_model)}")

    if weight is not None:
        output_path = f"{OUTPUT_PATH}model_{test}.pth"
        print(f"{output_path=}")
        new_model.store_weights_and_bias(output_path)

    print("======================\n")

    print("=======================")
    print("== Testing Inference ==")
    print("=======================")

    dataset: Dataset = select_dataset(old_model.dataset_name)(old_model)

    dataloader = list(dataset._actual_batch_generator(Dataset.Part.TRAIN))

    print("dataset:")
    print(dataset)

    if False:  # check_biases_shape
        # Both PyDTNN and PyTorch Biases have the same shape.
        print(" NEW MODEL LAYERS BIASES")
        for layer in new_model.layers:
            print(layer)
            print(f"layer.biases: {layer.biases.shape}")
        print(" OLD MODEL LAYERS BIASES")
        for layer in old_model.layers:
            print(layer)
            print(f"layer.biases: {layer.biases.shape}")

    # if TRAINING:
    #    print("OLD MODEL")
    #    pydtnn_training(model=old_model, dataset=dataset)
    #    print("NEW MODEL")
    #    pydtnn_training(model=new_model, dataset=dataset)

    if INFERENCE:
        if FIRST_PYTORCH:
            _pytorch_inference(pytorch_model, dataloader, kwargs, device)
            _pydtnn_inference(new_model, old_model, dataset)
        else:
            _pydtnn_inference(new_model, old_model, dataset, old_first=OLD_FIRST)
            _pytorch_inference(pytorch_model, dataloader, kwargs, device)


if __name__ == "__main__":
    main()



import unittest
import warnings

import numpy as np

from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.losses.loss import Loss, select as select_loss
from pydtnn.layers.layer import LayerError
from pydtnn.model import Model
from pydtnn.tests.abstract.common import verbose_test

from pydtnn.layers.layer import LayerBase
from pydtnn.tests.abstract.common import Params, TestCase
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils import print_with_header, random


class ModelCommonTestCase(TestCase):
    """
    Tests that two models with different parameters lead to the same results
    """

    # Compares results between an XX model {self.model1_desc} and other {self.model1_desc}
    model1_desc = "using A"
    model2_desc = "using B"

    rtol_default = 1e-4
    atol_default = 1e-5
    rtol_dict = {
        AdditionBlock: 1e-4,
        ConcatenationBlock: 1e-1,
        BatchNormalization: 1e-5,
        Conv2D: 1e-4,
    }
    atol_dict = {
        AdditionBlock: 5e-4,
        ConcatenationBlock: 1e-1,
        Conv2D: 1e-5,
        BatchNormalization: 1e-4,
    }

    def get_tolerance(self, layer: LayerBase) -> tuple[float, float]:
        rtol = self.rtol_default
        for cls, tol in self.rtol_dict.items():
            if isinstance(layer, cls):
                rtol = tol
                break

        atol = self.atol_default
        for cls, tol in self.atol_dict.items():
            if isinstance(layer, cls):
                atol = tol
                break

        # NOTE: Revise group layer tolerance
        if isinstance(layer, AbstractBlockLayer):
            for child in layer.children:
                crtol, catol = self.get_tolerance(child)
                rtol += crtol
                atol += catol

        return rtol, atol

    @staticmethod
    def get_model1_and_loss_func(model_name: str, overwrite_params: dict | None = None) -> tuple[Model, Loss]:
        # CPU model with no convGemm
        params = Params()
        # Begin of params configuration
        params.model_name = model_name  # type: ignore
        params.tensor_format = TensorFormat.NHWC.upper()
        # End of params configuration
        params_dict = vars(params)
        if overwrite_params is not None:
            params_dict.update(overwrite_params)
        try:
            model1 = Model(**params_dict)
        except LayerError as exc:
            raise unittest.SkipTest(f"Model {model_name} incompatible with {params_dict['dataset_name']}") from exc
        # loss function
        loss_func_name = model1.loss_func_name
        local_batch_size = model1.batch_size
        loss_cls = select_loss(loss_func_name)
        loss_func = loss_cls(shape=(local_batch_size, *model1.layers[-1].shape))
        loss_func.init_backend_from_model(model1)
        loss_func.initialize()
        return model1, loss_func

    def get_model2(self, model_name: str, overwrite_params: dict | None = None) -> Model:
        raise NotImplementedError()

    def copy_weights_and_biases(self, model1: Model, model2: Model):
        """
        Copy weights and biases from Model 1 to Model 2
        """
        model2.import_(model1)

    def get_first_dx(self, model: Model, loss_func: Loss, x: np.ndarray) -> np.ndarray:
        # random y target
        y_targ = np.asarray(random.random(x.shape), dtype=model.dtype, order='C', copy=True)
        # obtain first dx1
        global_batch_size = model.batch_size
        loss, dx = loss_func.compute(x, y_targ, global_batch_size)
        return dx

    def print_stats(self, x1: np.ndarray, x2: np.ndarray, rtol: float, atol: float) -> str:
        diff = x1 - x2
        return '\n' \
               f"\t{rtol=}\n"\
               f"\t{atol=}\n"\
               f"\tmax_diff={np.max(np.abs(diff))}\n" \
               f"\t{x1.max()=}\n" \
               f"\t{x2.max()=}\n" \
               f"\t{diff.max()=}\n" \
               f"\t{x1.min()=}\n" \
               f"\t{x2.min()=}\n" \
               f"\t{diff.min()=}\n" \
               f"\t{x1.std()=}\n" \
               f"\t{x2.std()=}\n" \
               f"\t{diff.std()=}\n" \
               f"\t{x1.mean()=}\n" \
               f"\t{x2.mean()=}\n" \
               f"\t{diff.mean()=}\n"
    # ---

    def do_model1_forward_pass(self, model1: Model, x0: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 1 forward pass
        """
        x1 = [x0[0]]
        for i, layer in enumerate(model1.layers):
            if verbose_test():
                print(layer)
            x1.append(layer.forward(np.asarray(x1[i], dtype=model1.dtype, order="C", copy=True)))
        return x1

    def do_model2_forward_pass(self, model2: Model, x1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 forward pass
        """
        x2 = [x1[0]]
        for i, layer in enumerate(model2.layers):
            if verbose_test():
                print(layer)
            x2.append(layer.forward(np.asarray(x1[i], dtype=model2.dtype, order="C", copy=True)))
        return x2

    @staticmethod
    def do_model1_backward_pass(model1: Model, dx0: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 1 backward pass
        """
        dx1 = [dx0[0]]
        for _, layer in reversed(list(enumerate(model1.layers))):
            if verbose_test():
                print(layer)
            dx1.insert(0, layer.backward(np.asarray(dx1[0], dtype=model1.dtype, order="C", copy=True)))
        return dx1

    def do_model2_backward_pass(self, model2: Model, dx1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 backward pass
        """
        dx2 = [dx1[-1]]
        for i, layer in reversed(list(enumerate(model2.layers))):
            if verbose_test():
                print(layer)
            dx2.insert(0, layer.backward(np.asarray(dx1[i + 1], dtype=model2.dtype, order="C", copy=True)))
        return dx2

    def compare_forward(self, model1: Model, x1: list[np.ndarray], model2: Model, x2: list[np.ndarray]):
        assert len(x1) == len(x2), "x1 and x2 should have the same length"
        if verbose_test():
            print()
            print(f"Comparing outputs of both models...")
        for i, layer in enumerate(model1.layers, 1):
            # Skip test on layers that behave randomly
            if not isinstance(layer, Dropout):
                rtol, atol = self.get_tolerance(layer)
                self.assertTrue(np.allclose(x1[i], x2[i], rtol=rtol, atol=atol),
                                f"Forward result from layers {layer.name_with_id} differ"
                                f" ({self.print_stats(x1[i], x2[i], rtol, atol)})")

    def compare_backward(self, model1: Model, dx1: list[np.ndarray], model2: Model, dx2: list[np.ndarray]):
        assert len(dx1) == len(dx2), f"dx1 and dx2 should have the same length {len(dx1)=}, {len(dx2)=}"
        if verbose_test():
            print("\nComparing outputs shapes.")
            min_dx = min(len(dx1), len(dx2))
            for i in range(min_dx):
                print(f"{i} - {dx1[i].shape=} ||\t{dx2[i].shape=}")
            for i in range(len(dx1) - len(dx2)):
                i = i + min_dx
                print(f"{i} - {dx1[i].shape=}")
            for i in range(len(dx2) - len(dx1)):
                i = i + min_dx
                print(f"{i} - {dx2[i].shape=}")

            print(f"\nComparing dx of both models...")
        for i, layer in enumerate(model2.layers, 0):
            # Skip test on layers that behave randomly
            if not isinstance(layer, Dropout):
                rtol, atol = self.get_tolerance(layer)
                if dx1[i].shape == dx2[i].shape:
                    allclose = np.allclose(dx1[i], dx2[i], rtol=rtol, atol=atol)
                else:
                    warnings.warn(f"dx shape on both models for {layer.name_with_id} differ:"
                                  f" [dx1.shape: {dx1[i].shape}, dx2.shape: {dx2[i].shape}]")
                    # Try flattening both
                    allclose = np.allclose(dx1[i].flatten(), dx2[i].flatten(), rtol=rtol, atol=atol)
                self.assertTrue(allclose,
                                f"Backward result from layer {layer.name_with_id} differ"
                                f" ({self.print_stats(dx1[i], dx2[i], rtol, atol)})")

    def do_test_model(self, model_name: str):
        """
        Compares results between a model that uses I2C and other that uses ConvGemm
        """

        # Model 1 forward
        model1, loss_func1 = self.get_model1_and_loss_func(model_name)
        model1.mode = Model.Mode.TRAIN

        model2 = self.get_model2(model_name)
        model2.mode = Model.Mode.TRAIN
        self.copy_weights_and_biases(model1, model2)

        x = [np.asarray(random.random((model1.batch_size, *model1.layers[0].shape)), dtype=model1.dtype, order='C', copy=True)]

        if verbose_test():
            print()
            print_with_header(f"Model {model1.model_name} 1 forward pass")
        x1 = self.do_model1_forward_pass(model1, x)

        x2 = x1.copy()

        # Model 2 forward
        if verbose_test():
            print_with_header(f"Model {model2.model_name} 2 forward pass")
        x2 = self.do_model2_forward_pass(model2, x2)

        # Compare forward results
        self.compare_forward(model1, x1, model2, x2)

        # Model 1 backward
        if verbose_test():
            print_with_header(f"Model {model1.model_name} 1 backward pass")
        dx = [self.get_first_dx(model1, loss_func1, x1[-1])]

        dx1 = self.do_model1_backward_pass(model1, dx)

        dx2 = dx1.copy()

        # Model 2 backward
        if verbose_test():
            print_with_header(f"Model {model2.model_name} 2 backward pass")
        dx2 = self.do_model2_backward_pass(model2, dx2)

        # Compare backward results
        self.compare_backward(model1, dx1, model2, dx2)

    def test_alexnet(self):
        f"""
        Compares results between an Alexnet model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("alexnet_cifar10")

    def test_vgg11(self):
        f"""
        Compares results between a VGG-11 BN model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("vgg11_cifar10")

    def test_vgg16bn(self):
        f"""
        Compares results between a VGG-16 BN model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("vgg16bn_cifar10")

    def test_resnet34(self):
        f"""
        Compares results between a Densenet model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("resnet34_cifar10")

    def test_densenet(self):
        f"""
        Compares results between a Densenet model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("densenet_cifar10")

    def test_simplecnn(self):
        f"""
        Compares results between a SimpleCNN model {self.model1_desc} and other {self.model1_desc}
        """
        self.do_test_model("simplecnn")



import numpy as np

from pydtnn.activations.log import Log
from pydtnn.backends.gpu.activations.activation import ActivationGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE

from pydtnn.libs import libcudnn as cudnn
import pycuda.gpuarray as gpuarray  # type: ignore
from pycuda.elementwise import ElementwiseKernel  # type: ignore


class LogGPU(Log[TensorGPU], ActivationGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log: ElementwiseKernel = None
        self.dlog: ElementwiseKernel = None

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.log = ElementwiseKernel(
            "{T} *in, {T} *out".format(T=DTYPE2CTYPE[self.model.dtype]),
            "out[i] = {func_log}(1.0 / (1.0 + {func_exp}(-in[i])));".format(
                func_log={np.float32: "logf", np.float64: "log"}[self.model.dtype], 
                func_exp={np.float32: "expf", np.float64: "exp"}[self.model.dtype]),
            "log_GPU")

        self.dlog = ElementwiseKernel(
            "{T} *in, {T} *out".format(T=DTYPE2CTYPE[self.model.dtype]),
            "out[i] = 1.0 / (1.0 + {func}(in[i]));".format(func={np.float32: "expf", np.float64: "exp"}[self.model.dtype]),
            "dlog_GPU")

        # Activations y
        y_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        self.log(x.ary, self.y.ary, stream=self.model.stream)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        # Compute dx
        self.dlog(dy.ary, self.dx.ary, stream=self.model.stream)
        return self.dx



from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
from pydtnn.layers.layer import LayerError
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape, Array

import numpy as np

CONCAT_DIM_NCHW = 1
CONCAT_DIM_NHWC = -1

class ConcatenationBlock[T: Array](AbstractBlockLayer[T]):
    def initialize_block_layer(self):
        super().initialize_block_layer()

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.concat_dim = CONCAT_DIM_NCHW
                if not all([tuple(o[CONCAT_DIM_NCHW:]) == tuple(self.out_shapes[0][CONCAT_DIM_NCHW:]) for o in self.out_shapes]):
                    raise LayerError(f"All output shape must have the same number of elements.\n{self.out_shapes}")
                self.out_co = [s[0] for s in self.out_shapes]
                self.idx_co = np.cumsum(self.out_co, axis=0)
                self.shape = (sum(self.out_co), *self.out_shapes[0][CONCAT_DIM_NCHW:])
            case TensorFormat.NHWC:
                self.concat_dim = CONCAT_DIM_NHWC
                if not all([tuple(o[:CONCAT_DIM_NHWC]) == tuple(self.out_shapes[0][:CONCAT_DIM_NHWC]) for o in self.out_shapes]):
                    raise LayerError(f"All output shape must have the same number of elements.\n{self.out_shapes}")
                self.out_co = [s[-1] for s in self.out_shapes]
                self.idx_co: np.ndarray = np.cumsum(self.out_co, axis=0)
                self.shape: ArrayShape = (*self.out_shapes[0][:CONCAT_DIM_NHWC], sum(self.out_co))
            case tensor_format:
                raise NotImplementedError(f"\"ConcatenationBlock\" is not implemented for \"{tensor_format}\" format.")
        self.co, self.ho, self.wo = self.model.decode_shape(self.shape)



from typing import Any
import numpy as np

import pycuda.driver as drv  # type: ignore
import pycuda.gpuarray as gpuarray  # type: ignore

from pydtnn.layers.fc import FC
from pydtnn.utils.performance_models import matmul_time
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.backends.gpu.utils import matmul_gpu, matvec_gpu
from pydtnn.utils.constants import ArrayShape, Parameters

class FCGPU(FC[TensorGPU], LayerGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.matmul = matmul_gpu
        self.matvec = matvec_gpu

    def _import_biases_db(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        
        cpu_ary = np.asarray(np.expand_dims(value, axis=0), dtype=self.model.dtype, order="C", copy=None)
        attribute.ary.set(cpu_ary)
        return
    # ---

    def _import_prop(self, key: str, value) -> None:
        match key:
            case Parameters.BIASES | Parameters.DB:
                return self._import_biases_db(key, value)
            # 
            case _:
                return super()._import_prop(key, value)
    # -----

    def _export_biases_db(self, key: str) -> Any:
        value = getattr(self, key)
        gpu_ary = value.ary
        cpu_ary = gpu_ary.get()

        return np.asarray(np.squeeze(cpu_ary, axis=0), dtype=np.float64, order="C", copy=True)
    # ---

    def _export_prop(self, key: str) -> Any:
        match key:
            case Parameters.BIASES | Parameters.DB:
                return self._export_biases_db(key)
            case _:
                return super()._export_prop(key)

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)
        self.stream_2 = drv.Stream()

        # Weights
        self.weights_cpu = self.weights_initializer(self.weights_shape, self.model.dtype)
        weights_gpu = gpuarray.to_gpu(self.weights_cpu)
        self.weights = TensorGPU(weights_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        if self.use_bias:
            # Biases
            self.biases_cpu = self.biases_initializer((1, *self.shape), self.model.dtype)
            biases_gpu = gpuarray.to_gpu(self.biases_cpu)
            self.biases = TensorGPU(biases_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        y_gpu = gpuarray.empty((self.model.batch_size, self.shape[0]), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        dx_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)
        self.dx.reshape((self.model.batch_size, *prev_shape))

        self.dw_cpu, self.dw = TensorGPU.initialize(self.weights.ary.shape, self.model.dtype,
                                                    tensor_format=self.model.tensor_format,
                                                    cudnn_dtype=self.model.cudnn_dtype,
                                                    gpudirect=self.model.gpudirect, 
                                                    drv=(drv if self.model.gpudirect else None))
        if self.use_bias:
            self.biases: TensorGPU
            self.db_cpu, self.db = TensorGPU.initialize(self.biases.ary.shape, self.model.dtype,
                                                        tensor_format=self.model.tensor_format,
                                                        cudnn_dtype=self.model.cudnn_dtype,
                                                        gpudirect=self.model.gpudirect, 
                                                        drv=(drv if self.model.gpudirect else None))

        self.one_vec_gpu = gpuarray.to_gpu(np.ones((self.model.batch_size,), self.model.dtype))
        self.nparams = self.weights.size + (self.biases.size if self.use_bias else 0)

        self.fwd_time = \
            matmul_time(m=self.model.batch_size, n=self.weights_cpu.shape[1], k=self.weights_cpu.shape[0],
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time = \
            matmul_time(m=self.weights_cpu.shape[0], n=self.weights_cpu.shape[1], k=self.model.batch_size,
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype) + \
            matmul_time(m=self.model.batch_size, n=self.weights_cpu.shape[0], k=self.weights_cpu.shape[1],
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)  # type: ignore (This is correct)

    def forward(self, x: TensorGPU) -> TensorGPU:
        m = x.ary.shape[0]
        n = ldb = ldc = self.weights.ary.shape[1]
        k = lda = x.ary.shape[1]
        trans_a, trans_b, alpha, beta = 'N', 'N', 1.0, 0.0

        # Compute a' = x @ weights
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                     self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUBLAS_MATMUL)
        self.matmul(self.model.cublas_handle, trans_b, trans_a, n, m, k, alpha,
                    self.weights.ary.gpudata, ldb,
                    x.ary.gpudata, lda, beta,
                    self.y.ary.gpudata, ldc, self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            alpha, beta = 1.0, 1.0
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                         self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN_SUM_BIASES)
            # Compute a = a' + biases
            cudnn.cudnnAddTensor(self.model.cudnn_handle, alpha, self.biases.desc,
                                 self.biases.ptr, beta, self.y.desc, self.y.ptr)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        # Compute dw
        m = lda = self.x.ary.shape[1]
        n = ldb = ldc = dy.ary.shape[1]
        k = dy.ary.shape[0]
        trans_a, trans_b, alpha, beta = 'T', 'N', 1.0, 0.0

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                     self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUBLAS_MATMUL_DW)
        self.matmul(self.model.cublas_handle, trans_b, trans_a, n, m, k, alpha,
                    dy.ary.gpudata, ldb, self.x.ary.gpudata, lda, beta,
                    self.dw.ptr_intp if self.model.gpudirect else self.dw.ary.gpudata, ldc, self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # DtoH dw when data parallelism and no GPU direct/NCCL is used
        if self.model.comm and not self.model.gpudirect and not self.model.enable_nccl:
            self.model.stream.synchronize()
            self.dw.ary.get_async(self.stream_2, self.dw_cpu)

        if self.use_bias:
            self.biases: TensorGPU
            # Compute db
            m = dy.ary.shape[0]
            n = lda = dy.ary.shape[1]
            trans_a, alpha, beta, inc_x, inc_y = 'N', 1.0, 0.0, 1, 1

            self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                         self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUBLAS_MATVEC_DB)
            self.matvec(self.model.cublas_handle, trans_a, n, m, alpha,
                        dy.ary.gpudata, lda, self.one_vec_gpu.gpudata, inc_x, beta,
                        self.db.ptr_intp if self.model.gpudirect else self.db.ary.gpudata,
                        inc_y, self.model.dtype)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

            # DtoH db when data parallelism and no GPU direct/NCCL is used
            if self.model.comm and not self.model.gpudirect and not self.model.enable_nccl:
                self.model.stream.synchronize()
                self.db.ary.get_async(self.stream_2, self.db_cpu)

        # Compute dx
        m = dy.ary.shape[0]
        n = ldc = self.weights.ary.shape[0]
        k = lda = ldb = dy.ary.shape[1]
        trans_a, trans_b, alpha, beta = 'N', 'T', 1.0, 0.0

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                     self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUBLAS_MATMUL_DX)
        self.matmul(self.model.cublas_handle, trans_b, trans_a, n, m, k, alpha,
                    self.weights.ary.gpudata, ldb,
                    dy.ary.gpudata, lda, beta,
                    self.dx.ary.gpudata, ldc, self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Convolutional layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any

# Functionality imports
from pydtnn.layers.conv_2d import Conv2D as _Conv2D
import pydtnn.converters.pytorch2pydtnn.common as cm

# ------------------ #


def Conv2d(args: Dict[str, Any]) -> _Conv2D:
    # https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html#torch.nn.Conv2d

    # PyTorch attributes:
    # Not used: in channels, device, dtype
    PYTORCH_KERNEL_SIZE = "kernel_size"  # INT or Tuple[INT, INT]
    PYTORCH_STRIDE = "stride"  # INT or Tuple[INT, INT]
    PYTORCH_PADDING = "padding"  # INT or Tuple[INT, INT]
    PYTORCH_DILATION = "dilation"  # INT
    PYTORCH_GROUPS = "groups"  # INT
    PYTORCH_BIAS = "bias"  # BOOL
    PYTORCH_OUPUT_CHANNELS = "out_channels"
    # PYTORCH_PADDING_MODE = "padding_mode" # STRING. Values: {"zeros", "reflect", "replicate", "circular"} | In PyDTNN "zeros" is the only implemented
    torch_dict_keys = [PYTORCH_KERNEL_SIZE, PYTORCH_STRIDE, PYTORCH_PADDING, PYTORCH_DILATION, PYTORCH_GROUPS, PYTORCH_BIAS, PYTORCH_OUPUT_CHANNELS]
    # ---- #

    # PyDTNN attributes:
    PYDTNN_FILTER_SHAPE = "filter_shape"
    PYDTNN_STRIDE = "stride"
    PYDTNN_PADDING = "padding"
    PYDTNN_DILATION = "dilation"
    PYDTNN_NFILTERS = "nfilters"
    PYDTNN_USE_BIAS = "use_bias"
    PYDTNN_NFILTERS = "nfilters"
    pydtnn_dict_keys = [PYDTNN_FILTER_SHAPE, PYDTNN_STRIDE, PYDTNN_PADDING, PYDTNN_DILATION, PYDTNN_NFILTERS, PYDTNN_USE_BIAS, PYDTNN_NFILTERS]
    # Not used: "activation" and "grouping"
    # Used, but in other place: "weights_initializer", "biases_initializer"
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    if PYDTNN_FILTER_SHAPE in layer_args:
        pool_shape = layer_args[PYDTNN_FILTER_SHAPE]
        if isinstance(pool_shape, int):
            layer_args[PYDTNN_FILTER_SHAPE] = (pool_shape, pool_shape)
        # else: It must be a Tuple[int, int], so it's okay
    # else: Nothing special

    return _Conv2D(**layer_args)



from collections.abc import Sequence

from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax


def simplemlp(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    _(Flatten())
    _(FC(shape=(512,), activation=Relu))
    _(FC(shape=(512,), activation=Relu))
    _(FC(shape=(512,), activation=Relu))
    _(FC(shape=output_shape, activation=Softmax))

    return model



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.categorical_mse import CategoricalMSE


class CategoricalMSECPU(CategoricalMSE[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        # return np.square(1 - y_pred[np.arange(b), np.argmax(y_targ, axis=1)]).mean()
        error = np.subtract(y_pred, y_targ, dtype=self.model.dtype)
        np.power(error, 2, out=error, dtype=self.model.dtype, casting="unsafe")
        return np.mean(error, dtype=self.model.dtype)




class MemoryCache(dict):
    """
    Dictionary derived class that can use the provided factory function to
    obtain a default value for a missing key. It differs from defaultdict in:

    * The provided factory function receives key as a parameter (which allows
      the generated value to depend on the given key).

    * If disable() is called, the instances of this class will clear their
      already stored values and will not store the next ones.

    """
    _preserve_values = True

    def __init__(self, default_factory=None, **kwargs):
        super().__init__(self, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self.default_factory(key)
            if self._preserve_values:
                self[key] = ret
            return ret

    @classmethod
    def disable(cls, update: bool = False):
        cls._preserve_values = False
        if not update:
            return
        import gc
        for obj in gc.get_objects():
            if isinstance(obj, cls):
                obj.clear()

    @classmethod
    def enable(cls):
        cls._preserve_values = True



from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.metrics.multiclass_confusion_matrix import MulticlassConfusionMatrix

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
import numpy as np
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore
from pydtnn.utils.constants import DTYPE2CTYPE

class MulticlassConfusionMatrixGPU(MulticlassConfusionMatrix[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()
        n = self.model.batch_size
        target_classes = self.model.output_shape[0]
        
        self.conf_matrix = TensorGPU.create_zeros_tensor(shape=(1, 1, target_classes, target_classes), dtype=np.dtype(np.int32), 
                                                        tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
        self.local_cm = TensorGPU.create_zeros_tensor(shape=(1, n, target_classes, target_classes), dtype=np.dtype(np.int32), 
                                                      tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
    # ----

    def __init_gpu_kernel__(self) -> Function:
        
        _name = "multiclass_confusion_matrix"
        code = """
        
        #define INDEX_FIRST_ONE_ON(y, var_class) for(i = 0; (i < num_classes) && !(y[i]); i++); var_class = i;
        #define SHIFT_POINTER_CM(p, i, j, num_classes) p + (i * num_classes + j)
        #define SHIFT_POINTER_LOCAL_CM(p, idx, i, j, num_i, num_j) p + ((idx * num_i + i) * num_j + j)
        
        __global__ void {name}({T} *y_targ, {T} *y_pred, int *cm, int *local_cm, const int num_classes, const int n)
        {{
            int idx, idx_i, i, j, target_class, predicted_class;

            const int base_idx = blockIdx.x * blockDim.x + threadIdx.x;
            const int workers = blockDim.x * gridDim.x;

            for(idx = base_idx; idx < n; idx += workers)
            {{
                INDEX_FIRST_ONE_ON(y_targ, target_class)
                INDEX_FIRST_ONE_ON(y_pred, predicted_class)

                (*(SHIFT_POINTER_LOCAL_CM(local_cm, idx, target_class, predicted_class, num_classes, num_classes))) += 1;
            }}
            
            // Accumulating the local values
            if (base_idx == 0)
            {{   
                for(idx_i = blockDim.x/2; idx_i > 0; idx_i >>= 1)
                {{
                    if(base_idx < idx_i)
                    {{
                        for(i = 0; i < num_classes; i++) for(j = 0; j < num_classes; j++)
                        {{
                            (*(SHIFT_POINTER_LOCAL_CM(local_cm, base_idx, i, j, num_classes, num_classes))) += (*(SHIFT_POINTER_LOCAL_CM(local_cm, base_idx + idx_i, i, j, num_classes, num_classes)));
                        }}
                    }}
                    __syncthreads();
                }}
            }}
            
            // Accumulating the local values into the output's tensor.
            if (base_idx == 0)
            {{
                for(i = 0; i < num_classes; i++) for(j = 0; j < num_classes; j++)
                {{
                    (*(SHIFT_POINTER_CM(cm, i, j, num_classes))) = (*(SHIFT_POINTER_LOCAL_CM(local_cm, base_idx, i, j, num_classes, num_classes)));
                }}
            }}
        }}
        """
        
        code = code.format(
            T = DTYPE2CTYPE[self.model.dtype],
            name = _name
        )
        module = SourceModule(code).get_function(_name)

        return module 
    #---


    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> np.ndarray:
        """
        The output will be a confusion matrix like this:
                |Predicted     |
        ________| 0  | 1  | 2  |
        Target|0| T0 | F1 | F2 |
              |1| F0 | T1 | F2 |
              |2| F0 | F1 | T2 |
        """

        n = y_pred.shape[0]
        target_classes = self.model.output_shape[0]

        self.conf_matrix.fill(0)
        self.local_cm.fill(0)

        n = np.int32(n)
        num_classes = np.int32(target_classes)

        self.kernel(y_targ.ary, y_pred.ary, 
                    self.conf_matrix.ary, self.local_cm.ary,
                    num_classes, n,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        return self.conf_matrix.ary.get()



from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
from pydtnn.utils.constants import Array


class Encoder[T: Array](AbstractBlockLayer[T]):
    def __init__(self, embedl: int = 64, d_k: int = 3, d_ff: int = 256, heads: int = 10, dropout_rate: float = 0.5):
        super().__init__()
        self.embedl = embedl
        self.heads = heads
        self.d_k = d_k
        self.d_ff = d_ff
        self.dropout_rate = dropout_rate
        self.y = self.dx = None



import os
import sys
import math
import string
import ctypes
import inspect
import weakref
import functools
import threading
from glob import glob
from queue import Queue
from importlib import import_module
from ctypes.util import find_library
from collections.abc import Iterable

import numpy as np


class BackgroundGenerator[T](threading.Thread):
    def __init__(self, generator: Iterable[T], max_prefetch=0):
        super().__init__()
        self.queue = Queue(max_prefetch)
        self.generator = generator
        self.daemon = True
        self.done = False
        self.start()

    def run(self):
        for item in self.generator:
            self.queue.put(item)
        self.queue.put(self)

    def __next__(self) -> T:
        if self.done:
            raise StopIteration()
        next_item = self.queue.get()
        self.done = next_item is self
        if self.done:
            raise StopIteration()
        return next_item

    def __iter__(self):
        return self


class Random:
    def __init__(self, seed=0) -> None:
        self._generators = weakref.WeakKeyDictionary[threading.Thread, np.random.Generator]()
        self.seed(seed)

    def seed(self, seed) -> None:
        self._seed = seed
        self._generators.clear()

    @property
    def _generator(self) -> np.random.Generator:
        thread = threading.current_thread()

        if thread not in self._generators:
            self._generators[thread] = np.random.default_rng(self._seed)

        return self._generators[thread]

    def __getattr__(self, key: str):
        return getattr(self._generator, key)


random: np.random.Generator = Random()  # type: ignore


def print_with_header(header, to_be_printed=None):
    print(f"# {header}")
    if to_be_printed is not None:
        print(to_be_printed)


def parse_bool(x):
    """Returns True if value is a user truthy value"""
    return str(x).lower() in {'true', '1', 'yes', 'y', 't'}


def get_attr_factory(o, name, factory):
    try:
        return getattr(o, name)
    except AttributeError:
        return factory()


def set_attr_default(o, name, value):
    try:
        return getattr(o, name)
    except AttributeError:
        setattr(o, name, value)
        return value


def set_attr_default_factory(o, name, factory):
    try:
        return getattr(o, name)
    except AttributeError:
        value = factory()
        setattr(o, name, value)
        return value

def load_library(name):
    """
    Loads an external library using ctypes.CDLL.

    It searches the library using ctypes.util.find_library(). If the library is
    not found, it traverses the LD_LIBRARY_PATH until it finds it. If it is not
    in any of the LD_LIBRARY_PATH paths, an ImportError exception is raised.

    Parameters
    ----------
    name : str
        The library name without any prefix like lib, suffix like .so, .dylib or
        version number (this is the form used for the posix linker option -l).

    Returns
    -------
    The loaded library.
    """
    path = find_library(name)
    if path is None:
        if sys.platform in ('linux2', 'linux'):
            full_name = f"lib{name}.so"
        elif sys.platform == 'darwin':
            full_name = f"lib{name}.dylib"
        elif sys.platform == 'win32':
            full_name = f"lib{name}.dll"
        else:
            raise NotImplementedError(f"Trying to load '{name}' library, but platform '{sys.platform}' is not yet supported!")

        for current_path in os.environ.get('LD_LIBRARY_PATH', '').split(':'):
            if os.path.exists(os.path.join(current_path, full_name)):
                path = os.path.join(current_path, full_name)
                break
        else:
            # Didn't find the library
            raise ImportError(f"Library '{full_name}' could not be found. Please add its path to LD_LIBRARY_PATH "
                              f"using 'export LD_LIBRARY_PATH={name.upper()}_LIB_PATH:$LD_LIBRARY_PATH' and "
                              f"then call this application again.")
    return ctypes.CDLL(path)


def blis():
    if not hasattr(blis, "lib"):
        blis.lib = load_library("blis")
    return blis.lib


def mkl():
    if not hasattr(mkl, "lib"):
        mkl.lib = load_library("mkl_rt")
    return mkl.lib


def convert_size(units: int, scale: int = 1000):
    size_name = ("", "K", "M", "G", "T", "P", "E", "Z", "Y")
    if units > 0:
        i = int(math.log(units, scale))
        p = math.pow(scale, i)
        s = round(units / p, 2)
    else:
        i = 0
        s = 0
    return f"{s}{size_name[i]}"


def convert_size_bytes(size_bytes):
    return f"{convert_size(size_bytes, scale=1024)}B"


# Matmul operation
# Warning: the output matrix can not be cached, as it will persist outside this method
def matmul(a: np.ndarray, b: np.ndarray, c: np.ndarray | None = None) -> np.ndarray:
    # if a.dtype == np.float32:
    #    c = slb.sgemm(1.0, a, b)
    # elif a.dtype == np.float64:
    #    c = slb.dgemm(1.0, a, b)
    # else:
    # Native numpy matmul gets more performance than scipy blas!
    if c is None:
        return a @ b
    else:
        return np.matmul(a, b, c)


def _matmul_xgemm(called_from, lib, a, b, c=None):
    order = 101  # 101 for row-major, 102 for column major data structures
    m = a.shape[0]
    n = b.shape[1]
    k = a.shape[1]
    if c is None:
        c = np.ones((m, n), a.dtype, order="C")
    # trans_{a,b} = 111 for no transpose, 112 for transpose, and 113 for conjugate transpose
    if a.flags["C_CONTIGUOUS"]:
        trans_a = 111
        lda = k
    elif a.flags["F_CONTIGUOUS"]:
        trans_a = 112
        lda = m
    else:
        raise ValueError(f"Matrix a data layout not supported by {called_from}().")
    if b.flags["C_CONTIGUOUS"]:
        trans_b = 111
        ldb = n
    elif b.flags["F_CONTIGUOUS"]:
        trans_b = 112
        ldb = k
    else:
        raise ValueError(f"Matrix b data layout not supported by {called_from}().")
    ldc = n
    alpha = 1.0
    beta = 0.0
    if a.dtype == np.float32:
        lib.cblas_sgemm(ctypes.c_int(order), ctypes.c_int(trans_a), ctypes.c_int(trans_b),
                        ctypes.c_int(m), ctypes.c_int(n), ctypes.c_int(k), ctypes.c_float(alpha),
                        ctypes.c_void_p(a.ctypes.data), ctypes.c_int(lda),
                        ctypes.c_void_p(b.ctypes.data), ctypes.c_int(ldb),
                        ctypes.c_float(beta), ctypes.c_void_p(c.ctypes.data), ctypes.c_int(ldc))
    elif a.dtype == np.float64:
        lib.cblas_dgemm(ctypes.c_int(order), ctypes.c_int(trans_a), ctypes.c_int(trans_b),
                        ctypes.c_int(m), ctypes.c_int(n), ctypes.c_int(k), ctypes.c_double(alpha),
                        ctypes.c_void_p(a.ctypes.data), ctypes.c_int(lda),
                        ctypes.c_void_p(b.ctypes.data), ctypes.c_int(ldb),
                        ctypes.c_double(beta), ctypes.c_void_p(c.ctypes.data), ctypes.c_int(ldc))
    else:
        raise TypeError(f"Type '{a.dtype}' not supported by {called_from}().")
    return c


def matmul_mkl(a, b, c=None):
    # os.environ['GOMP_CPU_AFFINITY'] = ""
    # os.environ['OMP_PLACES'] = ""
    return _matmul_xgemm("matmul_mkl", mkl(), a, b, c)


def matmul_blis(a, b, c=None):
    return _matmul_xgemm("matmul_blis", blis(), a, b, c)


def string_substitute(template, /, **mappings):
    """Shell-like opportunistic substitution"""
    return string.Template(template).safe_substitute(mappings)


def debug_line(*args):
    """Get line trace"""
    log = print

    frame_info = inspect.stack()[1]
    try:
        context = f"{frame_info.frame.f_globals["__name__"]}.{frame_info.function}:{frame_info.lineno}"
    finally:
        del frame_info

    log(f"{context} from {os.getpid()}:{threading.get_native_id()}", *args)


def debug_stack(*args, sep="|"):
    """Get stack trace"""
    log = print

    stack = inspect.stack()[1:]
    try:
        context = sep.join(
            f"{frame_info.frame.f_globals["__name__"]}.{frame_info.function}:{frame_info.lineno}"
            for frame_info in stack
        )
    finally:
        del stack

    log(f"{context} from {os.getpid()}:{threading.get_native_id()}", *args)


def debug_func(func):
    """Functions trace decorator"""
    log = print

    @functools.wraps(func)
    def wrapper(*args, **kwds):
        header = "DEBUG"
        frame_info = inspect.stack()[1]
        try:
            context = f"{func.__qualname__}{args!r}{kwds!r} from {frame_info.frame.f_globals["__name__"]}.{frame_info.function}:{frame_info.lineno} from {os.getpid()}:{threading.get_native_id()}"
        finally:
            del frame_info
        log(f"{header}: Call {context}")
        try:
            result = func(*args, **kwds)
        except BaseException as exc:
            log(f"{header}: Exc. {context} = {exc!r}")
            raise
        else:
            log(f"{header}: Ret. {context} = {result!r}")
            return result

    return wrapper


def find_component(package: str, name: str):
    def normalize(text: str) -> str:
        return text.lower().replace("_", "")

    try:
        module = import_module(f"{package}.{name}")
    except Exception as e:
        raise ValueError(f"{name!r} not found in {package!r}!") from e

    for attr in dir(module):
        if not attr.startswith("_") and normalize(name) == normalize(attr):
            return getattr(module, attr)
    else:
        raise ValueError(f"{name!r} not found in {module!r}!")



from typing import Self
from pydtnn.activations.activation import Activation
from pydtnn.utils.constants import ArrayShape, Array


class Relu[T: Array](Activation[T]):

    def __init__(self, shape: ArrayShape = (1,)):
        super().__init__(shape)
        # Will be initalized in "initialize"
        self.mask: T = None # type: ignore



from pydtnn.activations.leaky_relu import LeakyRelu
from pydtnn.utils.performance_models import im2col_time, col2im_time
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.backends.gpu.activations.activation import ActivationGPU
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE
import numpy as np
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_OPS_EVENT_enum

import pycuda.gpuarray as gpuarray  # type: ignore
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore


class LeakyReluGPU(LeakyRelu[TensorGPU], ActivationGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The following attributes will be initialized later.
        self.mask: TensorGPU = None # type: ignore
        self.y: TensorGPU = None # type: ignore
    
    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.cuda_fwd_func = self.cuda_adaptive_average_pooling_fwd(dtype=self.model.dtype)
        self.cuda_bwd_func = self.cuda_adaptive_average_pooling_bwd(dtype=self.model.dtype)

        self.total_num_threads = np.prod(self.grid, dtype=np.int32) * np.prod(self.block, dtype=np.int32)

        self.initialize_relu_2d_gpu(prev_shape)
    # ---
    
    def cuda_adaptive_average_pooling_fwd(self, dtype: np.dtype) -> Function:
        _func_name = "cuda_leaky_relu_fwd"
        _t = DTYPE2CTYPE[dtype]  # variable Type

        code = \
            """
__global__ void {func_name}({T}* x, {T}* max, {T}* mask,
                            float negative_slope, int num_workers, int N)
{{
    int idx, i;
    {T} elem;

    for(i = blockIdx.x * blockDim.x + threadIdx.x; i < N; i += num_workers)
    {{
        elem = x[i];

        if (elem > 0)
        {{
            max[i] = elem;
            mask[i] = 1;
        }}
        else if(elem < 0)
        {{
            max[i] = ({T}) (elem * negative_slope);
            mask[i] = negative_slope;
        }}
        else
        {{
            max[i] = 0;
            mask[i] = 0;
        }}
    }}
}}
"""
        code = code.format(func_name=_func_name, T=_t)

        return SourceModule(code).get_function(_func_name)
    # -----
    
    def cuda_adaptive_average_pooling_bwd(self, dtype: np.dtype) -> Function:
        _func_name = "cuda_leaky_relu_bwd"
        _t = DTYPE2CTYPE[dtype]  # variable Type

        code = \
            """
__global__ void {func_name}({T}* dx, {T}* dy, {T}* mask,
                            int num_workers, int N)
{{
    int i;

    for(i = blockIdx.x * blockDim.x + threadIdx.x; i < N; i += num_workers)
        dx[i] = dy[i] * mask[i];
}}
"""
        code = code.format(func_name=_func_name, T=_t)

        return SourceModule(code).get_function(_func_name)
    # -----
    
    def forward(self, x: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)

        n = np.prod(x.shape, dtype=np.int32)

        self.cuda_fwd_func(x.ary, self.mask.ary, self.max.ary,
                           np.float32(self.negative_slope), self.total_num_threads, n,
                           grid=self.grid, block=self.block, stream=self.model.stream)

        self.y: TensorGPU = self.mask

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        return self.y
    
    def backward(self, dy: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)

        n = np.prod(dy.shape, dtype=np.int32)

        self.cuda_bwd_func(self.dx.ary, dy.ary, self.mask.ary,
                           self.total_num_threads, n,
                           grid=self.grid, block=self.block,
                           stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
        return self.dx
    
    def initialize_relu_2d_gpu(self, prev_shape: ArrayShape) -> None:
        self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)
        self.shape = prev_shape

        n: int = self.model.batch_size * self.hi * self.wi * self.ci

        _max = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.max = TensorGPU(_max, self.model.tensor_format, self.model.cudnn_dtype)

        _mask = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.mask = TensorGPU(_mask, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.fwd_time = \
            im2col_time(m=self.ci, n=n, cpu_speed=self.model.cpu_speed,
                        memory_bw=self.model.memory_bw, dtype=self.model.dtype)
        self.bwd_time = \
            col2im_time(m=self.ci, n=n, cpu_speed=self.model.cpu_speed,
                        memory_bw=self.model.memory_bw, dtype=self.model.dtype)
    


import numpy as np

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.conv_2d import Conv2DGPU, MACROS_NCHW, MACROS_NHWC
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE

from pydtnn.utils.tensor import TensorFormat, format_transpose
from typing import Any, override

import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.driver import Function  #type: ignore

class Conv2DPointwiseGPU(Conv2DGPU):

    def _initializing_special_parameters(self):
        self.kh = self.kw = 1
        # Setting weights
        match self.model.tensor_format:
                case TensorFormat.NCHW:
                    self.weights_shape = (self.co, self.ci)
                case TensorFormat.NHWC:
                    self.weights_shape = (self.co, self.ci)
                case _:
                    raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        #--

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        func_name: str = ""
        macros: str = ""
        self.bias_sum_bwd: Function = None

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                func_name = "cuda_depthwise_conv_2d_{fwd_bwd}_nchw"
                macros = MACROS_NCHW
                self.bias_sum_bwd = self.cuda_sum_bias_axis_023()
            case TensorFormat.NHWC:
                func_name = "cuda_depthwise_conv_2d_{fwd_bwd}_nhwc"
                macros = MACROS_NHWC
                self.bias_sum_bwd = self.cuda_sum_bias_axis_012()
            case _:
                raise NotImplementedError(f"\"conv_2d_gpu_depthwise\" is not implemented for \"{self.model.tensor_format}\" format.")

        self.total_num_threads = np.prod(self.grid, dtype=np.int32) * np.prod(self.block, dtype=np.int32)

        y_gpu = gpuarray.to_gpu(np.zeros(shape=(self.model.batch_size, *self.shape), dtype=self.model.dtype))
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        dx_gpu = gpuarray.zeros((self.model.batch_size, *self.shape), self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.forward = self._forward_pointwise
        self.backward = self._backward_pointwise
        self.fwd_func: Function = self.cuda_pointwise_conv_2d_fwd(func_name.format(fwd_bwd="fwd"), macros)
        self.bwd_func: Function = self.cuda_pointwise_conv_2d_bwd(func_name.format(fwd_bwd="bwd"), macros)
        self.bias_sum_fwd: Function = self.cuda_bias_pointwise_conv_2d_fwd("bias_pointwise_conv_2d_fwd", macros)
    # ----

    def _forward_pointwise(self, x: TensorGPU) -> TensorGPU:

        self.x = x
        self.y.fill(0)

        n, c, h, w = self.model.decode_shape(x.shape)  # type: ignore (it's okay)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        self.fwd_func(x.ary, self.weights.ary, self.y.ary,
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.co), self.total_num_threads,
                      grid=self.grid, block=self.block, stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN_SUM_BIASES)
            self.bias_sum_fwd(x.ary, self.biases.ary,
                              np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                              np.int32(n * c * h * w),
                              self.total_num_threads,
                              grid=self.grid, block=self.block, stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.y
    # -----

    def _backward_pointwise(self, dy: TensorGPU) -> TensorGPU:
        n, c, h, w = self.model.decode_shape(dy.shape)  # type: ignore (it's okay)
        self.dx.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        self.fwd_func(dy.ary, self.x.ary, self.weights.ary,
                      self.dx.ary, self.dw.ary,
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.kh), np.int32(self.kw), np.int32(self.ho), np.int32(self.wo),
                      self.total_num_threads,
                      grid=self.grid, block=self.block, stream=self.model.stream)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DB)
            self.bias_sum_bwd(dy.ary, self.db.ary,
                              np.int32(c), np.int32(h), np.int32(w),
                              np.int32(n * c * h * w), self.total_num_threads,
                              grid=self.grid, block=self.block, stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.dx
    # -----

    @override
    def _export_weights_dw(self, key: str) -> Any:
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: self.ci, self.co
                # NCHW's dst: self.co, self.ci
                gpu_ary = value.ary
                cpu_ary = gpu_ary.get()
                return np.asarray(format_transpose(np.squeeze(cpu_ary, axis=(1, 2)), "IO", "OI"), dtype=np.float64, order="C", copy=True)
            case TensorFormat.NCHW:
                # NHWC's src: self.ci, self.co
                # NCHW's dst: self.co, self.ci
                gpu_ary = value.ary
                cpu_ary = gpu_ary.get()
                return np.asarray(np.squeeze(cpu_ary, axis=(2, 3)), dtype=np.float64, order="C", copy=True)
            case default:
                return super()._export_prop(key)
    # ------

    @override
    def _import_weights_dw(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: self.ci, self.co
                # NCHW's dst: self.co, self.ci
                cpu_ary = np.asarray(np.expand_dims(format_transpose(value, "OI", "IO"), axis=(1, 2)), dtype=self.model.dtype, order="C", copy=None)
                attribute.ary.set(cpu_ary)
                return
            case TensorFormat.NCHW:
                gpu_ary = attribute.ary
                cpu_ary = np.asarray(np.expand_dims(value, axis=(2, 3)), dtype=self.model.dtype, order="C", copy=None)
                gpu_ary.set(cpu_ary)
                return 
            case default:
                return super()._export_prop(key)
    # ---

#########################################################################################################
## CUDA CODE ##
###############
    def cuda_pointwise_conv_2d_fwd(self, _func_name: str, _macros: str) -> Function:

        code = \
"""
{macros}

__global__ void {func_name}({T}* x, {T}* k, {T}* y,
                            int n, int c, int h, int w,
                            int yc, int num_workers)
{{
    int idx, ni, ci, hi, wi, yci;
    {T} val_k, val_x;

    int N = n*c*h*w;

    // k.shape = (yc, x's c)

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        ni = INDEX_N(idx, N, n);
        ci = INDEX_C(idx, c, h, w);
        hi = INDEX_H(idx, c, h, w);
        wi = INDEX_W(idx, c, h, w);

        val_x = *(SHIFT_POINTER(x, c, h, w, ni, ci, hi, wi));
        for(yci = 0; yci < yc; yci++)
        {{
            //y = x * k
            //val_k = k[yci][ci]; ==> val_k = k + (yci * c + ci);
            //val_k = k[ci][yci]; ==> val_k = k + (ci * kc + yci);
            val_k = *(SHIFT_POINTER_K(k, c, yc, ci, yci));
            *(SHIFT_POINTER(y, yc, h, w, ni, yci, hi, wi)) += ({T}) (val_x * val_k);
        }}
    }}
}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = code.format(macros=_macros,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ---

    def cuda_pointwise_conv_2d_bwd(self, _func_name: str, _macros: str) -> Function:

        code = \
"""
{macros}

__global__ void {func_name}({T}* dy, {T}* x, {T}* k,
                            {T}* dx, {T}* dw,
                            int n, int c, int h, int w,
                            int xc, int num_workers)
{{
    int idx, ni, ci, hi, wi, xci;
    {T} val_dy, val_k, val_x;

    int N = n*c*h*w;

    // NCHW: k.shape = dw.shape = (dy's c , x's c)
    // NHWC: k.shape = dw.shape = (x's c, dy's c)

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        ni = INDEX_N(idx, N, n);
        ci = INDEX_C(idx, c, h, w);
        hi = INDEX_H(idx, c, h, w);
        wi = INDEX_W(idx, c, h, w);

        val_dy = *SHIFT_POINTER(dy, c, h, w, ni, ci, hi, wi);
        for(xci = 0; xci < kc; xci++)
        {{
            //dw = x * dy
            val_x = *(SHIFT_POINTER(x, xc, h, w, ni, xci, hi, wi));
            *(SHIFT_POINTER_K(dw, c, xc, ci, xci)) = ({T}) (val_x * val_dy);

            //dx = w * dy
            val_k = *(SHIFT_POINTER_K(k, c, xc, ci, xci));
            *(SHIFT_POINTER(dx, kc, h, w, nn, xci, hi, wi)) += ({T}) (val_k * val_dy);
        }}
    }}
}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = code.format(macros=_macros,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ---

    def cuda_bias_pointwise_conv_2d_fwd(self, _func_name: str, _macros: str) -> Function:

        code = \
"""
{macros}

__global__ void {func_name}({T}* y, {T}* b,
                            int n, int c, int h, int w,
                            int N,
                            int num_workers)
{{
    int idx, ni, ci, hi, wi;

    // self.biases.shape = (self.co,)

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        ni = INDEX_N(idx, N, n);
        ci = INDEX_C(idx, c, h, w);
        hi = INDEX_H(idx, c, h, w);
        wi = INDEX_W(idx, c, h, w);

        *(SHIFT_POINTER(y, c, h, w, ni, ci, hi, wi)) += (*(b+ci));
    }}
}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = code.format(macros=_macros,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ---
#########################################################################################################



import numpy as np

import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.elementwise import ElementwiseKernel  #type: ignore

from pydtnn.backends.gpu.layers.abstract.block_layer import AbstractBlockLayerGPU
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import DTYPE2CTYPE, ArrayShape


class ConcatenationBlockGPU(ConcatenationBlock[TensorGPU], AbstractBlockLayerGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.concat: ElementwiseKernel = None
        self.split: ElementwiseKernel = None
        self.dy: list[TensorGPU] = None  #type: ignore
        self.idx_co = None  #type: ignore

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)
        # @warning: super().initialize() calls self.initialize_block_layer() (don't call it again)
        self.concat = ElementwiseKernel(
            "{T} *dst, {T} *src, int N, int H, int W, int C, int first_c, int last_c".format(T=DTYPE2CTYPE[self.model.dtype]),
            {TensorFormat.NHWC:
                """int c_ = i % C;
                   if (first_c <= c_ && c_ < last_c) {
                       int w_ = i / C % W;
                       int h_ = i / (W*C) % H;
                       int n_ = i / (H*W*C) % N;
                       int i_ = n_ * H * W * (last_c-first_c) + h_ * W * (last_c-first_c) + w_ * (last_c-first_c) + (c_-first_c);
                       dst[i] = src[i_];
                   }
                """,
             TensorFormat.NCHW:
                """int c_ = i / (H*W) % C;
                   if (first_c <= c_ && c_ < last_c) {
                       int w_ = i % W;
                       int h_ = i / W % H;
                       int n_ = i / (C*H*W) % N;
                       int i_ = n_ * (last_c-first_c) * H * W + (c_-first_c) * H * W + h_ * W + w_;
                       dst[i] = src[i_];
                   }
                """}[self.model.tensor_format],
            "concat")

        self.split = ElementwiseKernel(
            "{T} *src, {T} *dst, int N, int H, int W, int C, int first_c, int last_c".format(T=DTYPE2CTYPE[self.model.dtype]),
            {TensorFormat.NHWC:
                """int c_ = i % C;
                   if (first_c <= c_ && c_ < last_c) {
                       int w_ = i / C % W;
                       int h_ = i / (W*C) % H;
                       int n_ = i / (H*W*C) % N;
                       int i_ = n_ * H * W * (last_c-first_c) + h_ * W * (last_c-first_c) + w_ * (last_c-first_c) + (c_-first_c);
                       dst[i_] = src[i];
                   }
                """,
             TensorFormat.NCHW:
                """int c_ = i / (H*W) % C;
                   if (first_c <= c_ && c_ < last_c) {
                       int w_ = i % W;
                       int h_ = i / W % H;
                       int n_ = i / (C*H*W) % N;
                       int i_ = n_ * (last_c-first_c) * H * W + (c_-first_c) * H * W + h_ * W + w_;
                       dst[i_] = src[i];
                   }
                """}[self.model.tensor_format],
            "split")

        # Activations y
        y_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)
        # Derivative dy
        self.dy = []
        for i, p in enumerate(self.paths):
            dy_gpu = gpuarray.empty((self.model.batch_size, *self.out_shapes[i]), self.model.dtype)
            self.dy.append(TensorGPU(dy_gpu, self.model.tensor_format, self.model.cudnn_dtype))

    def forward(self, x: TensorGPU) -> TensorGPU:
        for i, p in enumerate(self.paths):
            y_i = x
            for layer in p:
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                y_i = layer.forward(y_i)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONCAT)
            self.concat(self.y.ary, y_i.ary, self.model.batch_size, self.ho, self.wo, self.co,
                        0 if i == 0 else self.idx_co[i - 1], self.idx_co[i])
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        for i, p in enumerate(self.paths):
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SPLIT)
            self.split(dy.ary, self.dy[i].ary, self.model.batch_size, self.ho, self.wo, self.co,
                       0 if i == 0 else self.idx_co[i - 1], self.idx_co[i])
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
            dx_i = self.dy[i]
            for layer in reversed(p):
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
                dx_i = layer.backward(dx_i)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            if i == 0:
                self.dx = dx_i
            else:
                alpha, beta = 1.0, 1.0
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                             self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ELTW_SUM)
                cudnn.cudnnAddTensor(self.model.cudnn_handle, alpha, dx_i.desc,
                                     dx_i.ptr, beta, self.dx.desc, self.dx.ptr)
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



import numpy as np

from pydtnn.backends.cpu.optimizers.optimizer import OptimizerCPU
from pydtnn.optimizers.rmsprop import RMSProp

from pydtnn.backends.cpu.layers.layer import LayerCPU


class RMSPropCPU(RMSProp[np.ndarray], OptimizerCPU):

    def initialize(self, list_layers: list[LayerCPU]) -> None:

        for layer in list_layers:
            list_grad_vars = list(layer.grad_vars.keys())

            if len(list_grad_vars) != 0:
                self.context[layer.id] = dict[str, np.ndarray]()  # type: ignore 
                for w_ in list_grad_vars:
                    w: np.ndarray = getattr(layer, w_)
                    self.context[layer.id]["cache_%s" % w_] = np.zeros_like(w, dtype=layer.model.dtype, order="C")

    def update(self, layer: LayerCPU) -> None:
        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            cache: np.ndarray = self.context[layer.id]["cache_%s" % w_]  # type: ignore 
            w: np.ndarray
            dw: np.ndarray

            if not (self.are_all_zeros(w) and self.are_all_zeros(dw) and self.are_all_zeros(cache)):
                # NOTE: The operations are unrolled in order to reduce the memory consumed by intermediate copies of the variables during the operations.

                # cache = self.rho * cache + (1 - self.rho) * dw ** 2
                np.multiply(cache, self.rho, out=cache, 
                            dtype=self.dtype)
                temp_dw = np.power(dw, 2, dtype=self.dtype, order="C")
                np.multiply(temp_dw, (1 - self.rho), out=temp_dw, 
                        dtype=self.dtype)
                np.add(cache, temp_dw, out=cache,
                       dtype=self.dtype)
                
                # w -= self.learning_rate * (self.decay * w + (dw / np.sqrt(cache + self.epsilon)))
                temp_w = np.multiply((self.learning_rate * self.decay), w, dtype=self.dtype, order="C")
                np.subtract(w, temp_w, out=w, 
                            dtype=self.dtype)
                del temp_w

                temp_cache = np.add(cache, self.epsilon, dtype=self.dtype, order="C")
                np.sqrt(temp_cache, out=temp_cache,
                        dtype=self.dtype)
                temp_dw = np.divide(dw, temp_cache, dtype=self.dtype, order="C")
                del temp_cache

                np.multiply(temp_dw, self.learning_rate, out=temp_dw,
                            dtype=self.dtype)
                np.subtract(w, temp_dw, out=w,
                            dtype=self.dtype)
                del temp_dw
            # else: continue



import numpy as np

from pydtnn.metrics.regression_mse import RegressionMSE
from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore

from pydtnn.utils.constants import DTYPE2CTYPE

class RegressionMSEGPU(RegressionMSE[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()

        n = self.model.batch_size
        num_classes = self.model.output_shape

        self.res = TensorGPU.create_zeros_tensor(shape=(1, ), dtype=np.dtype(self.model.dtype), 
                                                 tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)

        self.local_res = TensorGPU.create_zeros_tensor(shape=(n, *num_classes), dtype=np.dtype(self.model.dtype), 
                                                       tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
    # ----

    def __init_gpu_kernel__(self) -> Function:
        _name = "regression_mse"
        code = """

        #define SHIFT_2D_AR(p, i, j, dim_j) (p + ((i * dim_j) + j))

        __global__ void {name} ({T} *y_targ, {T} *y_pred, {T} *res, {T} *local_res, int n, int labels)
        {{
            int i, idx;
            {T} diff, val_targ, val_pred;
        
            int base_idx = blockIdx.x * blockDim.x + threadIdx.x;
            int workers = blockDim.x * gridDim.x;

            for(idx = base_idx; idx < n; idx += workers)
            {{
                *(local_res + idx) = ({T}) 0.0;

                for(i = 0; i < labels; i++)
                {{
                    // val_targ = y_targ[idx][i];
                    val_targ = (*SHIFT_2D_AR(y_targ, idx, i, labels));

                    // val_pred = y_pred[idx][i];
                    val_pred = (*SHIFT_2D_AR(y_pred, idx, i, labels));
                    
                    diff = val_targ - val_pred;
                    *(local_res + idx) += (diff * diff);
                }}
                
            }}

            if(base_idx == 0)
            {{
                (*res) = (*local_res);
                for(idx = 1; (idx < n); idx++)
                    (*res) += (*(local_res + idx));

                (*res) /= (n * labels);
            }}
        }}
        """.format(T=DTYPE2CTYPE[self.model.dtype],
                   name=_name)
        
        module = SourceModule(code).get_function(_name)
        return module

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> float:
        n = y_pred.shape[0]
        num_classes = y_pred.shape[1]

        self.res.fill(0)
        self.local_res.fill(0)

        n = np.int32(n)
        num_classes = np.int32(num_classes) 
        self.kernel(y_targ.ary, y_pred.ary, 
                    self.res.ary, self.local_res.ary,
                    n, num_classes,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        return self.res.ary.get()[0]



import numpy as np
from pydtnn.backends.cpu.utils.average_pool_2d_nchw_cython import average_pool_2d_bwd_nchw_cython, average_pool_2d_fwd_nchw_cython
from pydtnn.backends.cpu.utils.average_pool_2d_nhwc_cython import average_pool_2d_bwd_nhwc_cython, average_pool_2d_fwd_nhwc_cython
from pydtnn.backends.cpu.utils.im2col_1ch_nchw_cython import col2im_1ch_nchw_cython, im2col_1ch_nchw_cython
from pydtnn.backends.cpu.utils.im2row_1ch_nhwc_cython import im2row_1ch_nhwc_cython, row2im_1ch_nhwc_cython
from pydtnn.backends.cpu.layers.abstract.pool_2d_layer import AbstractPool2DLayerCPU
from pydtnn.layers.average_pool_2d import AveragePool2D

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum


class AveragePool2DCPU(AveragePool2D[np.ndarray], AbstractPool2DLayerCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.y: np.ndarray

    def initialize(self, prev_shape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)
        y_shape = self.model.encode_shape((self.model.batch_size, self.co, self.ho, self.wo))
        # NOTE: This attribute only stores data, its value before the operation doesn't matter; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self.y = np.zeros(y_shape, dtype=self.model.dtype, order="C")


    def _forward_nhwc_i2c(self, x: np.ndarray) -> np.ndarray:

        x_rows = np.zeros((x.shape[0] * self.ci * self.ho * self.wo, self.kh * self.kw), dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        im2row_1ch_nhwc_cython(x, x_rows,
                               self.kh, self.kw, self.ho, self.wo,
                               self.vpadding, self.hpadding,
                               self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        y: np.ndarray = np.mean(x_rows, axis=1, dtype=self.model.dtype)
        return y.reshape((-1, self.ho, self.wo, self.co), order="C", copy=None)

    def _forward_nhwc_cython(self, x: np.ndarray) -> np.ndarray:

        y = self.y[:x.shape[0], :]
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        average_pool_2d_fwd_nhwc_cython(x, y,
                                        self.kh, self.kw, self.ho, self.wo,
                                        self.vpadding, self.hpadding,
                                        self.vstride, self.hstride,
                                        self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nchw_i2c(self, x: np.ndarray) -> np.ndarray:
        n, c, _, _ = x.shape
        x_cols = np.zeros((self.kh * self.kw, n * c * self.ho * self.wo), dtype=self.model.dtype, order="C")

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        im2col_1ch_nchw_cython(x, x_cols,
                               self.kh, self.kw, self.ho, self.wo,
                               self.vpadding, self.hpadding,
                               self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        y: np.ndarray = np.mean(x_cols, axis=1, dtype=self.model.dtype)
        return y.reshape((-1, self.co, self.ho, self.wo), order="C", copy=None)

    def _forward_nchw_cython(self, x: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        y = self.y[:x.shape[0], :]
        average_pool_2d_fwd_nchw_cython(x, y,
                                        self.kh, self.kw, self.ho, self.wo,
                                        self.vpadding, self.hpadding,
                                        self.vstride, self.hstride,
                                        self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward_nhwc_i2c(self, dy: np.ndarray) -> np.ndarray:
        pool_size = np.prod(self.pool_shape)
        dy_rows = np.tile(dy.reshape(-1, 1, copy=False) / pool_size, (1, pool_size))  # type: ignore (it is correct.)
        dx = np.zeros_like(dy, dtype=self.model.dtype)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        row2im_1ch_nhwc_cython(dy_rows, dx,
                               dy.shape[0], self.hi, self.wi, self.ci,
                               self.kh, self.kw, self.ho, self.wo,
                               self.vpadding, self.hpadding,
                               self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return dx.reshape((-1, self.hi, self.wi, self.ci), order="C", copy=None)

    def _backward_nhwc_cython(self, dy: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        # NOTE: It's necessary a new zero-initalized "dx" in every call since may be some values that are not re-set in the cython's function.
        dx = np.zeros((dy.shape[0], self.hi, self.wi, self.ci), dtype=self.model.dtype)
        average_pool_2d_bwd_nhwc_cython(dy, dx,
                                        dy.shape[0], self.hi, self.wi, self.ci,
                                        self.kh, self.kw, self.ho, self.wo,
                                        self.vpadding, self.hpadding,
                                        self.vstride, self.hstride,
                                        self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)

    def _backward_nchw_i2c(self, dy: np.ndarray) -> np.ndarray:
        pool_size = np.prod(self.pool_shape)
        dy_cols = np.tile(dy.flatten() / pool_size, (pool_size, 1))  # type: ignore (it is correct.)
        dy_cols = np.asarray(dy_cols, dtype=self.model.dtype, order="C", copy=None)
        dx = np.zeros((dy.shape[0], self.hi, self.wi, self.ci), dtype=self.model.dtype, order="C")

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        col2im_1ch_nchw_cython(dy_cols, dx,
                               dy.shape[0], self.hi, self.wi, self.ci,
                               self.kh, self.kw, self.ho, self.wo,
                               self.vpadding, self.hpadding,
                               self.vstride, self.hstride, 
                               self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return dx.reshape((-1, self.ci, self.hi, self.wi), order="C", copy=None)

    def _backward_nchw_cython(self, dy: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        # NOTE: It's necessary a new zero-initalized "dx" in every call since may be some values that are not re-set in the cython's function.
        dx = np.zeros((dy.shape[0], self.ci, self.hi, self.wi), dtype=self.model.dtype, order="C")
        average_pool_2d_bwd_nchw_cython(dy, dx,
                                        dy.shape[0], self.hi, self.wi, self.ci,
                                        self.kh, self.kw, self.ho, self.wo,
                                        self.vpadding, self.hpadding,
                                        self.vstride, self.hstride,
                                        self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)



import warnings

import numpy as np

from pydtnn.backends.cpu.optimizers.optimizer import OptimizerCPU
from pydtnn.backends.cpu.utils.oktopk_utils_cython import compute_dense_acc_cython, intersect_2d_indexes_cython, reset_residuals_cython, update_sparsed_weights_cython, update_sparsed_weights_mv_cython
from pydtnn.layer_base import LayerBase
from pydtnn.optimizers.oktopk import OkTopk
from pydtnn.utils.sparse import SparseMatrixCOO

try:
    from pydtnn.libs.libmpi import MPI
except (ImportError, ModuleNotFoundError):
    pass


class OkTopkCPU(OkTopk[np.ndarray], OptimizerCPU):
    def initialize(self, list_layers: list[LayerBase]) -> None:
        super().initialize(list_layers)

        for layer in list_layers:
            self.iterations[layer.id] = 0
            self.all_local_th[layer.id] = {dw_: None for dw_ in layer.grad_vars.values()}
            self.all_global_th[layer.id] = {dw_: None for dw_ in layer.grad_vars.values()}
            self.all_residuals[layer.id] = {dw_: None for dw_ in layer.grad_vars.values()}
            self.all_boundaries[layer.id] = {dw_: None for dw_ in layer.grad_vars.values()}

    def update(self, layer):
        for w_, dw_ in layer.grad_vars.items():
            # Get layer weights and gradients
            w, dw = getattr(layer, w_), getattr(layer, dw_)

            # Reshape dw to 2D matrix
            self.dw_original_shape = dw.shape
            if len(self.dw_original_shape) != 2:
                dw = dw.reshape(dw.shape[0], -1)
            self.dw_2d_shape = dw.shape

            # Compute k from: layer_params * self.density
            k = int(np.prod(self.dw_original_shape) * self.density)
            k = self.min_k_layer if k < self.min_k_layer else k

            # Initialize current layer-parameter values
            self.local_th = self.all_local_th[layer.id][dw_]
            self.global_th = self.all_global_th[layer.id][dw_]
            self.boundaries = self.all_boundaries[layer.id][dw_]
            if self.all_residuals[layer.id][dw_] is None:
                self.all_residuals[layer.id][dw_] = np.zeros_like(dw, dtype=layer.model.dtype)

            # Compute acc
            acc = self._compute_acc(self.all_residuals[layer.id][dw_], dw, self.learning_rate)

            # Main part of ok-topk: compute the values that contribute to the update and its indexes
            coo_u, indexes = self._ok_sparse_allreduce(acc, self.iterations[layer.id], k, self.tau, self.tau_prime)

            # Update residuals
            self.all_residuals[layer.id][dw_] = self._reset_residuals(acc, indexes)

            # Save for next updates thresholds and boundaries
            self.all_local_th[layer.id][dw_] = self.local_th
            self.all_global_th[layer.id][dw_] = self.global_th
            self.all_boundaries[layer.id][dw_] = self.boundaries

            # Perform the weights update
            self._update_weights(layer, w_, w, coo_u)

        self.iterations[layer.id] += 1

    def _compute_acc(self, residuals, dw, learning_rate, method="cython"):
        """
        Compute acc, where: acc = residuals + (learning_rate * dw)

        Parameters:
            residuals (np.array): 2D dense matrix with the current layer residuals
            dw (np.array): 2D dense matrix with the current layer gradients
            learning_rate (float): learning rate float value
            method (string, optional): The method to use for updating the weights. It can be 'cython' or 'numpy'. Default is 'cython'.

        Warning:
            'cython' method does not provide the same exact accuracy as 'numpy'.

        Returns:
            acc (np.array): 2D dense matrix with the updated residuals
        """

        self._show_message_only_once(f"\n\nIn '_compute_acc', the method that it is being used is '{method}'")

        if method == "cython":
            return compute_dense_acc_cython(residuals, dw, learning_rate)

        if method == "numpy":
            return residuals + (learning_rate * dw)

        raise NotImplementedError(f"Method '{method}' not implemented")

    def _reset_residuals(self, acc, indexes, method="cython"):
        """
        Update residuals: set zero value if it is in indexes, else acc value is set.
        If density is 100% and some gradients are zero, scipy will be removing those indexes even if no sparsity is applied.
        Thus, to simulate 100% density, residuals must be always zero.
        This means that a slightly sparse factor will may remove more values because the gradients are already zero.

        Parameters:
            acc (np.array): 2D dense matrix
            indexes (tuple(np.array, np.array)): a tuple with rows and cols
            method (string, optional): The method to use for updating the weights. It can be 'cython' or 'numpy'. Default is 'cython'.

        Returns:
            residuals (np.array): which is the same as acc with the values in indexes set to zero.
        """

        self._show_message_only_once(f"In '_reset_residuals', the method that it is being used is '{method}'")

        if self.density == 1:
            return np.zeros_like(acc)

        if method == "cython":
            assert (self._has_canonical_format(indexes))
            return reset_residuals_cython(acc, indexes[0], indexes[1])

        if method == "numpy":
            if len(indexes[0]) > 0:
                acc[indexes] = 0
            return acc

        raise NotImplementedError(f"Method '{method}' not implemented")

    def _update_weights(self, layer, w_type, w, coo_u, method="cython"):
        """
        Update weights: w -= (u / self.model.nprocs)
        and set to weight layer attribute: setattr(layer, w_type, w)

        Parameters:
            layer (int): layer id
            w_type (string): weight param type (bias, weight, ...)
            w (np.array): N dimensional dense weights matrix/tensor
            coo_u (SparseMatrixCOO): Sparse 2D gradient matrix in COO format to update w
            method (string, optional): The method to use for updating the weights. It can be 'cython' or 'numpy'. Default is 'cython'.

        Returns:
            (void): instead it directly applies the result to the weight layer attribute
        """

        self._show_message_only_once(f"In '_update_weights', the method that it is being used is '{method}'")

        if method == "cython":
            if len(self.dw_original_shape) != 2:
                w = w.reshape(w.shape[0], -1)
            w = update_sparsed_weights_cython(w, coo_u.data, coo_u.row, coo_u.col)
            if len(self.dw_original_shape) != 2:
                w = w.reshape(self.dw_original_shape)
            setattr(layer, w_type, w)
            return

        if method == "cython_with_vel_and_momentum":
            if self.momentum == 0:
                warnings.warn("If momentum is 0 use 'cython' method, it produces the same output but it is faster", RuntimeWarning)

            if len(self.dw_original_shape) != 2:
                w = w.reshape(w.shape[0], -1)
            velocity = getattr(layer, "velocity_%s" % w_type, np.zeros_like(w, dtype=layer.model.dtype))
            w, velocity = update_sparsed_weights_mv_cython(w, coo_u.data, coo_u.row, coo_u.col, velocity, self.momentum)
            if len(self.dw_original_shape) != 2:
                w = w.reshape(self.dw_original_shape)
            setattr(layer, w_type, w)
            setattr(layer, "velocity_%s" % w_type, velocity)
            return

        if method == "numpy":
            if len(self.dw_original_shape) != 2:
                w = w.reshape(w.shape[0], -1)
            w[coo_u.row, coo_u.col] -= coo_u.data
            if len(self.dw_original_shape) != 2:
                w = w.reshape(self.dw_original_shape)
            setattr(layer, w_type, w)
            return

        if method == "numpy_with_vel_and_momentum":
            if self.momentum == 0:
                warnings.warn("If momentum is 0 use just 'numpy' method, it produces the same output but it is faster", RuntimeWarning)

            if len(self.dw_original_shape) != 2:
                w = w.reshape(w.shape[0], -1)
            velocity = getattr(layer, "velocity_%s" % w_type, np.zeros_like(w, dtype=layer.model.dtype))
            velocity *= self.momentum
            velocity[coo_u.row, coo_u.col] += coo_u.data
            w[coo_u.row, coo_u.col] -= velocity[coo_u.row, coo_u.col]
            if len(self.dw_original_shape) != 2:
                w = w.reshape(self.dw_original_shape)
            setattr(layer, w_type, w)
            setattr(layer, "velocity_%s" % w_type, velocity)
            return

        if method == "like_sgd":
            """Use only for debugging purposes"""
            warnings.warn("This method should be used only in case of debugging for performance reasons.", RuntimeWarning)

            dw = coo_u.toarray()
            if len(self.dw_original_shape) != 2:
                dw = dw.reshape(self.dw_original_shape)
            velocity = getattr(layer, "velocity_%s" % w_type, np.zeros_like(w, dtype=layer.model.dtype))
            velocity = self.momentum * velocity + dw
            w -= velocity  # Oktopk already computes acc with learning_rate
            setattr(layer, w_type, w)
            setattr(layer, "velocity_%s" % w_type, velocity)
            return

        raise NotImplementedError(f"Method '{method}' not implemented")

    def _ok_sparse_allreduce(self, acc, t, k, space_repartition_t, thresholds_re_evaluation_t):
        """
        Performs the Ok-Topk sparse allreduce operation.
        This method executes the Ok-Topk sparse allreduce algorithm, which
        optimizes communication by only exchanging the most significant
        gradient values (top-k) across distributed processes. The method
        periodically re-evaluates the thresholds and repartitions the
        gradient space to maintain efficiency and accuracy.

        Parameters:
            acc (np.array): 2D dense gradient matrix accumulation values.
            t (int): Current iteration number.
            k (int): Number of top-k gradient values to select in the current layer.
            space_repartition_t (int): Interval of iterations for space repartitioning.
            thresholds_re_evaluation_t (int): Interval of iterations for threshold re-evaluation.

        Returns:
            out (tuple with two elements:):
                - coo_u (SparseMatrixCOO): The updated gradient values in 2D sparse format.
                - indexes (tuple(np.array, np.array)): The indices of the top-k gradient values that were updated.
        """

        if t % thresholds_re_evaluation_t == 0:
            self.local_th = self._th_re_evaluate(acc, k, input_format="dense")

        if t % space_repartition_t == 0:
            self.boundaries = self._space_repartition(acc, self.local_th)

        coo_reduced_region_topk, local_topk_indexes = self._split_and_reduce(acc, self.local_th, self.boundaries)

        if t % thresholds_re_evaluation_t == 0:
            coo_all_reduced_topk = self._allgather(coo_reduced_region_topk)
            self.global_th = self._th_re_evaluate(coo_all_reduced_topk, k, input_format="coo")

        coo_u, global_topk_indexes = self._balance_and_allgather(coo_reduced_region_topk, self.global_th)
        indexes = self._intersect_indexes(local_topk_indexes, global_topk_indexes)
        return coo_u, indexes

    def _th_re_evaluate(self, matrix, k, input_format=None, method="numpy_sort"):
        """
        Return the absolute gradient threshold for a given matrix.

        Parameters:
            matrix (np.array or SparseMatrixCOO): A 2D gradient matrix, in np.array for 'dense' input_format or SparseMatrixCOO for 'coo' input_format.
            k (int): Indicating the number of top gradient values to consider.
            input_format (string): Either 'dense' for a dense matrix or 'coo' for a sparse matrix in COO format.
            method (string, optional): The method to use for threshold selection. It can be 'numpy_sort' or 'numpy_partition'.

        Returns:
            threshold (float): The absolute gradient threshold based on the top k values.
        """

        self._show_message_only_once(f"In '_th_re_evaluate', the method that it is being used is '{method}'")

        if k <= 0:
            return 0.0

        if input_format == "coo" and matrix.nnz == 0:
            return 1.0

        if input_format == "dense" and method == "numpy_sort":
            sorted_matrix = np.sort(np.abs(matrix).flatten())
            threshold = sorted_matrix[max(-k, -len(sorted_matrix))]
            return threshold

        if input_format == "coo" and method == "numpy_sort":
            sorted_data = np.sort(np.abs(matrix.data))
            threshold = sorted_data[max(-k, -len(sorted_data))]
            return threshold

        if input_format == "dense" and method == "numpy_partition":
            flat_matrix = np.abs(matrix).flatten()
            if k > len(flat_matrix):
                return flat_matrix.min()
            threshold = np.partition(flat_matrix, -k)[-k]
            return threshold

        if input_format == "coo" and method == "numpy_partition":
            flat_matrix = np.abs(matrix.data)
            if k > len(flat_matrix):
                return flat_matrix.min()
            threshold = np.partition(flat_matrix, -k)[-k]
            return threshold

        raise NotImplementedError(f"Method '{method}' with format '{input_format}' not implemented")

    def _space_repartition(self, acc, local_th, balanced=True):
        """
        Returns the boundaries of the regions of the gradient matrix for the split and reduce phase.

        Parameters:
            acc (np.array): 2D dense gradient matrix values
            local_th (float): local process gradient threshold
            balanced (boolean, optional): if not balanced a static row partition is performed,
                                          if balanced a topk gradiend distribution is considered in the row partition

        Warning:
            Balanced space repartition does not provide the same exact accuracy as static space repartition.

        Returns:
            boundaries (np.array): [row_end_p0, row_end_p1, row_end_p2, ...]
        """

        self._show_message_only_once(f"In '_space_repartition', balanced = '{balanced}' is being used")

        if not balanced:
            boundaries = np.zeros(self.model.nprocs, dtype=np.int32)
            total_rows = self.dw_original_shape[0]
            block_size = total_rows // self.model.nprocs
            for i in range(0, self.model.nprocs - 1):
                boundaries[i] = block_size * (i + 1)
            boundaries[self.model.nprocs - 1] = total_rows
            return boundaries

        if balanced:
            coo_topk = SparseMatrixCOO.from_dense_top_selection(acc, local_th)

            current_row = 0
            current_proc = 0
            rows = coo_topk.row
            topk_in_current_proc = 0
            total_rows = coo_topk.shape[0]
            boundaries = np.zeros(self.model.nprocs, dtype=np.int32)
            topk_per_proc = coo_topk.nnz // self.model.nprocs
            topk_per_row = np.zeros(total_rows, dtype=np.int32)
            np.add.at(topk_per_row, rows, 1)

            while current_proc < self.model.nprocs - 1:
                if current_row < total_rows:
                    topk_in_current_proc += topk_per_row[current_row]
                    if topk_in_current_proc >= topk_per_proc:
                        boundaries[current_proc] = current_row
                        topk_in_current_proc = 0
                        current_proc += 1
                    current_row += 1
                else:
                    boundaries[current_proc] = current_row
                    current_proc += 1
            boundaries[self.model.nprocs - 1] = total_rows

            global_boundaries = self.model.comm.allreduce(boundaries, op=MPI.SUM) // self.model.nprocs

            return global_boundaries

    def _split_and_reduce(self, acc, local_th, boundaries):
        """
        First main phase of ok_sparse_allreduce.
        Split the gradients into partitions and reduce them by selecting top-k values.
        Each worker receives sparse regions from the other workers and and then conducts the reduction locally.

        Parameters:
            acc (np.arrray): 2D gradient matrix accumulation values in dense format.
            local_th (float): Local threshold for selecting top-k values.
            boundaries (np.array): Boundaries for partitioning the gradient space like [row_end_p0, row_end_p1, row_end_p2, ...]

        Returns:
            out (tuple with two elements:):
                - coo_reduced_region_topk (SparseMatrixCOO): The reduced top-k gradient values in COO format.
                - local_topk_indexes (tuple(np.array, np.array)): The indices of the top-k gradient values selected locally.
        """

        coo_topk = SparseMatrixCOO.from_dense_top_selection(acc, local_th)
        coo_reduced_region_topk = self._reduce_topk(coo_topk, boundaries)
        return coo_reduced_region_topk, coo_topk.get_indexes()

    def _balance_and_allgather(self, coo_reduced_region_topk, global_th):
        """
        Second main phase of ok_sparse_allreduce.
        Performs the allgather of the coo_reduced_region_topk values among workers.

        Parameters:
            coo_reduced_region_topk (SparseMatrixCOO): a 2D sparse gradient matrix.
            global_th (float): the global threshold to perfrom top selection.

        Returns:
            out (tuple with two elements:):
                - coo_allgather_topk (SparseMatrixCOO): A 2D sparse gradient matrix with the global top-k selection.
                - reduced_region_global_topk_indexes (tuple(np.array, np.array)): The indices of the top-k gradient values region reduced.
        """

        # 1. Global topk selection
        coo_reduced_region_global_topk = coo_reduced_region_topk.top_selection(global_th, inplace=False)

        # 2. Data packaging
        # TODO

        # 3. Data balancing
        # TODO

        # 4. Allgatherv using recursive doubling
        coo_allgather_topk = self._allgather(coo_reduced_region_global_topk)
        return coo_allgather_topk, coo_reduced_region_global_topk.get_indexes()

    def _intersect_indexes(self, local_indexes, global_indexes):
        """
        Calculates the intersection of two sets of indices of 2D.
        The assertion statement is only executed when the script is not run in optimized mode (python3 -O script.py).
        Remember that '_has_canonical_format' should only be used for debugging/development purposes to assert that indexes are correct.
        Indexes in scipy are usually in canonical format, so it should not be necessary to evaluate the indexes format.
        When optimized mode is enabled (python3 -O script.py), the assert sentences are not computed.

        Parameters:
            local_indexes (tuple(np.array, np.array)): a tuple of two numpy arrays representing row and column indices, sorted by rows, then by columns.
            global_indexes (tuple(np.array, np.array)): a tuple of two numpy arrays representing row and column indices, sorted by rows, then by columns.

        Returns:
            intersected_indexes (tuple(np.array, np.array)): Set of tuples representing the common indices.

        Example:
            - local_indexes  = (np.array([0, 1, 2, 3, 3, 4]) , np.array([4, 6, 5, 1, 7, 3]))
            - global_indexes = (np.array([0, 1, 3, 3, 3]), np.array([1, 6, 1, 5, 7]))
            - output: (array([1, 3, 3]), array([6, 1, 7]))
        """

        local_rows, local_cols = local_indexes
        global_rows, global_cols = global_indexes
        return intersect_2d_indexes_cython(local_rows, local_cols, global_rows, global_cols)

    def _reduce_topk(self, coo_topk, boundaries, method="p2p_region_wise_reduce_destination_rotation_and_bucketing"):
        """
        Reduce the topk elements in regions defined by boundaries.

        Parameters:
            coo_topk (SparseMatrixCOO): a 2D sparse array in COO format with the values and indexes of topk.
            boundaries (np.array): boundaries for partitioning the gradient space like [row_end_p0, row_end_p1, row_end_p2, ...].
            method (str, optional): The method to use for reduce topk

        Returns:
            coo_reduced_region (SparseMatrixCOO): The reduced topk values in COO format.
        """
        self._show_message_only_once(f"In '_reduce_topk', the method that it is being used is '{method}'")

        if self.model.nprocs == 1:
            return coo_topk

        if method == "collective_allreduce_then_slice":
            warnings.warn("This reduce_topk method ('collective_allreduce_then_slice') should be used only in case of debugging for performance reasons.", RuntimeWarning)
            all_reduced_coo = self.model.comm.allreduce(coo_topk, op=MPI.SUM)
            row_start = 0 if self.model.rank == 0 else boundaries[self.model.rank - 1]
            row_end = boundaries[self.model.rank]
            return all_reduced_coo.slice(row_start, row_end)

        if method == "collective_region_wise_reduce_sync":
            row_start = 0
            reduced_regions_coo = [None] * self.model.nprocs
            for region in range(self.model.nprocs):
                row_end = boundaries[region]
                reduced_regions_coo[region] = self.model.comm.reduce(coo_topk.slice(row_start, row_end), op=MPI.SUM, root=region)
                row_start = row_end
            return reduced_regions_coo[self.model.rank]

        if method == "collective_region_wise_reduce_async":
            """It is not possible with the current mpi4py version to generate a buffer with indexes and values and operate with them"""
            pass

        if method == "p2p_region_wise_reduce_static_destination":
            # Prepare a vector region for storing the partial sums
            coo_region_partial_sum = [None] * self.model.nprocs
            for region in range(self.model.nprocs):
                row_start = 0 if region == 0 else boundaries[region - 1]
                row_end = boundaries[region]
                coo_region_partial_sum[region] = coo_topk.slice(row_start, row_end)

            # Overlaps comm. steps with computation (sparse sum)
            # On comm_step i: P{rank} sends to P{rank + 1} region{rank - i % nprocs}.
            destination = (self.model.rank + 1) % self.model.nprocs
            receive_from = (self.model.rank - 1) % self.model.nprocs
            for comm_step in range(1, self.model.nprocs):
                region_to_send = (self.model.rank - comm_step) % self.model.nprocs
                region_to_recv = (self.model.rank - comm_step - 1) % self.model.nprocs
                # recv_req = self.model.comm.irecv(source=receive_from)
                # self.model.comm.send(coo_region_partial_sum[region_to_send], dest=destination)
                # coo_region_partial_sum[region_to_recv] += recv_req.wait()
                coo_region_partial_sum[region_to_recv] += self.model.comm.sendrecv(coo_region_partial_sum[region_to_send],
                                                                             dest=destination, source=receive_from)

            return coo_region_partial_sum[self.model.rank]

        if method == "p2p_region_wise_reduce_destination_rotation_and_bucketing":
            # There are (nprocs - 1) messages to send (excluding self)
            total_sends = self.model.nprocs - 1
            requests = [None] * total_sends

            # Compute local slice of coo_topk (the "self" region)
            row_start = 0 if self.model.rank == 0 else boundaries[self.model.rank - 1]
            row_end = boundaries[self.model.rank]
            coo_reduced_region = coo_topk.slice(row_start, row_end)

            # Process sends and receives in buckets.
            bucket_size = 2
            region = (self.model.rank + 1) % self.model.nprocs
            for comm_step in range(0, total_sends, bucket_size):
                # The current bucket may have fewer messages than bucket_size (i.e. the last bucket)
                current_bucket_size = min(bucket_size, total_sends - comm_step)
                # Non-blocking sends for the current bucket
                for i in range(current_bucket_size):
                    row_start = 0 if region == 0 else boundaries[region - 1]
                    row_end = boundaries[region]
                    requests[comm_step + i] = self.model.comm.isend(coo_topk.slice(row_start, row_end), dest=region)
                    region = (region + 1) % self.model.nprocs
                # After sending the bucket, perform the receives sequentially for the same bucket.
                for i in range(current_bucket_size):
                    coo_reduced_region += self.model.comm.recv()

            MPI.Request.Waitall(requests)
            return coo_reduced_region

        raise NotImplementedError(f"Method '{method}' not implemented")

    def _allgather(self, local_data, input_format="SparseMatrixCOO"):
        """
        Gathers data from all processes.

        Parameters:
            local_data (np.ndarray or SparseMatrixCOO): The local data to be gathered.
            input_format (str, optional): The format of the input data.
        Returns:
            gathered_data (np.ndarray or SparseMatrixCOO): The gathered global data in the specified format.
        """

        if self.model.nprocs == 1:
            return local_data

        if input_format == "SparseMatrixCOO":
            gathered = self.model.comm.allgather(local_data.get_triplet())
            all_val = np.concatenate([t[0] for t in gathered])
            all_row = np.concatenate([t[1] for t in gathered])
            all_col = np.concatenate([t[2] for t in gathered])
            return SparseMatrixCOO(all_val, all_row, all_col, self.dw_2d_shape, has_canonical_format=True)

        if input_format == "dense":
            warnings.warn("Try to avoid dense communications!", RuntimeWarning)
            return np.concatenate(self.model.comm.allgather(local_data))

        raise NotImplementedError(f"Input format '{input_format}' not implemented")

    def _show_message_only_once(self, message):
        """
        Show information messages only once to assess the selected functions are being used.

        Parameters:
            message (str): The message to show.
        Returns:
            void (None):
        """
        if self.model.rank == 0:
            if message not in self.info_messages:
                self.info_messages.add(message)
                print(message)

    def _has_canonical_format(self, indexes):
        """
        Check if indexes follows the COO canonical format:
            - Indexes are sorted by row and then by column
            - There are no duplicate entries
        This function is computationally expensive and therefore should only be used for developing/debugging purposes.
        This function should only be used in developement to assert that sparse matrices have canonical format.

        Parameters:
            indexes (tuple(np.ndarray, np.ndarray)): indexes to check

        Returns:
            has_canonical_format (bool): True if indexes are in canonical format, False if not.
        """

        warnings.warn("This function ('has_canonical_format') should be used only in case of debugging for performance reasons.", RuntimeWarning)

        row, col = indexes

        if len(row) != len(col):
            raise AssertionError("'row' and 'col' must have the same length")

        if len(row) == 0:
            return True

        if not np.all(row[:-1] <= row[1:]):
            return False

        for i in range(len(row) - 1):
            if row[i] == row[i + 1] and col[i] >= col[i + 1]:
                return False
        return True



import numpy as np

from pydtnn.backends.cpu.utils.relu_cython import relu_cython
from pydtnn.activations.relu import Relu
from pydtnn.backends.cpu.activations.activation import ActivationCPU
from pydtnn.utils.constants import ArrayShape


class ReluCPU(Relu[np.ndarray], ActivationCPU):

    def __init__(self, shape: ArrayShape = (1,)):
        super().__init__(shape)
        self.mask: np.ndarray = None # type: ignore (will be initalized in "initialize")

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        # NOTE: These attributes only store data, their value before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros((self.model.batch_size, *self.prev_shape), dtype=self.model.dtype, order="C")
        self._mask = np.zeros((self.model.batch_size, *self.prev_shape), dtype=np.int8, order="C")
        self.dx = np.zeros((self.model.batch_size, *self.prev_shape), dtype=self.model.dtype, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:
        n = x.shape[0]
        self.y = self._y[:n, :]
        self.mask = self._mask[:n, :]
        relu_cython(x.reshape(-1, copy=False, order="C"), self.y.reshape(-1, copy=False, order="C"), self.mask.reshape(-1, copy=False, order="C"))
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        dx = self.dx[:dy.shape[0], :]
        np.multiply(dy, self.mask, out=dx, dtype=self.model.dtype, order="C")
        return dx



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def adaptive_avg_pooling_fwd_nchw_cython[T: _npDT](x: _npDT_4Dims[T], pooled_x: _npDT_4Dims[T]) -> None:
    """
    Args:
        x (npDT_4Dims): data input.
        pooled_x (npDT_4Dims): ndarray where the output will be stored. It must be filled with zeros.
    Returns:
        Nothing; the return is stored in "dx".
    """


def adaptive_avg_pooling_bwd_nchw_cython[T: _npDT](dy: _npDT_4Dims[T], dx: _npDT_4Dims[T]) -> None:
    """
    Args:
        dy (npDT_4Dims): data input.
        dx (npDT_4Dims): ndarray where the output will be stored. It must be filled with zeros.
    Returns:
        Nothing; the return is stored in "dx".
    """



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydtnn.activations.activation import Activation
from pydtnn.layers.layer import Layer
from pydtnn.utils.initializers import InitializerFunc, glorot_uniform, zeros
from pydtnn.utils.constants import Array, ArrayShape, Parameters


class FC[T: Array](Layer[T]):

    def __init__(self, shape: ArrayShape = (1,),
                 activation: "type[Activation] | None" = None,
                 use_bias=True,
                 weights_initializer: InitializerFunc = glorot_uniform,
                 biases_initializer: InitializerFunc = zeros):
        super().__init__(shape)
        self.act = activation
        self.use_bias = use_bias
        self.weights_initializer = weights_initializer
        self.biases_initializer = biases_initializer
        self.grad_vars = {Parameters.WEIGHTS: Parameters.DW}
        if self.use_bias:
            self.grad_vars[Parameters.BIASES] = Parameters.DB

    def initialize(self, prev_shape: ArrayShape, x: T | None) -> None:
        super().initialize(prev_shape, x)
        self.weights_shape = (*prev_shape, *self.shape)



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def im2col_1ch_nchw_cython[T: _npDT](x: _npDT_4Dims[T],
                                     cols: _npDT_2Dims[T],
                                     kh: int, kw: int, ho: int, wo: int,
                                     vpadding: int, hpadding: int,
                                     vstride: int, hstride: int,
                                     vdilation: int, hdilation: int) -> None:
    """
    Args:
        x (npDT_4Dims): The 4 dimensional array (the image).
        cols (npDT_2Dims): The 2 dimensional array where the image as columns is stored.
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.
    Returns:
        Nothing, it is stored in \"cols\".
    """


def col2im_1ch_nchw_cython[T: _npDT](cols: _npDT_2Dims[T],
                                     x: _npDT_4Dims[T],
                                     n: int, h: int, w: int, c: int,
                                     kh: int, kw: int, ho: int, wo: int,
                                     vpadding: int, hpadding: int,
                                     vstride: int, hstride: int,
                                     vdilation: int, hdilation: int) -> None:
    """
    Args:
        cols (npDT_2Dims): The 2 dimensional array.
        x (npDT_4Dims): The 4 dimensional array where the output image is stored.
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): vertical padding value.
        hpadding (int): horizontal padding value.
        vstride (int): vertical stride value.
        hstride (int): horizontal stride value.
        vdilation (int): vertical dilation value.
        hdilation (int): horizontal dilation value.
    Returns:
        Nothing, it is stored in \"x\".
    """



from pydtnn.layers.layer import Layer
from pydtnn.utils.constants import Array, ArrayShape


class Scalar[T: Array](Layer[T]):
    def __init__(self, shape: ArrayShape = (1,), scale: float = 1.0):
        super().__init__(shape)
        self.scale = scale
